from datetime import datetime
from uuid import uuid4

import polars as pl
from pytz import UTC

# Define the DataFrame DestinationRecord model
destination_record_schema = pl.Schema(
    [
        # Bizon system information
        ("bizon_id", str),
        ("bizon_extracted_at", pl.Datetime(time_unit="us", time_zone="UTC")),
        ("bizon_loaded_at", pl.Datetime(time_unit="us", time_zone="UTC")),
        # Source record information
        ("source_record_id", str),
        ("source_timestamp", pl.Datetime(time_unit="us", time_zone="UTC")),
        ("source_data", str),
    ]
)


def transform_to_df_destination_records(df_source_records: pl.DataFrame, extracted_at: datetime) -> pl.DataFrame:
    """Return a Polars DataFrame from a list of DestinationRecord objects"""
    return df_source_records.select(
        pl.Series([uuid4().hex for _ in range(df_source_records.height)]).alias("bizon_id"),
        pl.Series([extracted_at for _ in range(df_source_records.height)]).alias("bizon_extracted_at"),
        pl.Series([datetime.now(tz=UTC) for _ in range(df_source_records.height)]).alias("bizon_loaded_at"),
        pl.col("id").alias("source_record_id"),
        pl.col("timestamp").alias("source_timestamp"),
        pl.col("data").alias("source_data"),
    )



from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Optional, Tuple, Union

import polars as pl
from loguru import logger
from pydantic import BaseModel, Field

from bizon.common.models import SyncMetadata
from bizon.engine.backend.backend import AbstractBackend
from bizon.engine.backend.models import JobStatus
from bizon.source.callback import AbstractSourceCallback
from bizon.source.config import SourceSyncModes

from .buffer import DestinationBuffer
from .config import (
    AbstractDestinationConfig,
    AbstractDestinationDetailsConfig,
    DestinationTypes,
)
from .models import transform_to_df_destination_records


class DestinationBufferStatus(str, Enum):
    """Destination buffer status"""

    RECORDS_WRITTEN = "RECORDS_WRITTEN"
    RECORDS_WRITTEN_THEN_BUFFERED = "RECORDS_WRITTEN_THEN_BUFFERED"
    RECORDS_BUFFERED = "RECORDS_BUFFERED"
    NO_RECORDS = "NO_RECORDS"


class DestinationIteration(BaseModel):
    success: bool = Field(..., description="Success status of the iteration")
    error_message: Optional[str] = Field(None, description="Error message if iteration failed")
    records_written: int = Field(0, description="Number of records written to the destination")
    from_source_iteration: Optional[int] = Field(None, description="From source iteration identifier buffer starts")
    to_source_iteration: Optional[int] = Field(
        None, description="To source iteration identifier buffer ends, inclusive"
    )
    pagination: Optional[dict] = Field(None, description="Source pagination for next interation recovery purposes")


class AbstractDestination(ABC):

    def __init__(
        self,
        sync_metadata: SyncMetadata,
        config: AbstractDestinationDetailsConfig,
        backend: AbstractBackend,
        source_callback: AbstractSourceCallback,
    ):
        self.sync_metadata = sync_metadata
        self.config = config
        self.backend = backend
        self.buffer = DestinationBuffer(
            buffer_size=self.config.buffer_size, buffer_flush_timeout=self.config.buffer_flush_timeout
        )
        self.source_callback = source_callback
        self.destination_id = config.destination_id

        self._record_schemas = None
        self._clustering_keys = None

    @property
    def record_schemas(self):
        if self._record_schemas is None and self.config.record_schemas:
            self._record_schemas = {
                schema.destination_id: schema.record_schema for schema in self.config.record_schemas
            }
        return self._record_schemas

    @property
    def clustering_keys(self):
        if self._clustering_keys is None and self.config.record_schemas:
            self._clustering_keys = {
                schema.destination_id: schema.clustering_keys for schema in self.config.record_schemas
            }
        return self._clustering_keys

    @abstractmethod
    def check_connection(self) -> bool:
        pass

    @abstractmethod
    def write_records(self, df_destination_records: pl.DataFrame) -> Tuple[bool, Union[str, None]]:
        """Write records to destination for the given iteration and return success status and error message"""
        pass

    def finalize(self) -> bool:
        """Finalize destination after writing records, usually used to copy data from temp table to main table"""
        pass

    def buffer_flush_handler(self, session=None) -> DestinationIteration:
        # TODO: Add retry strategy

        # Initialize destination iteration
        destination_iteration = DestinationIteration(
            success=False,
            records_written=0,
            pagination=self.buffer.pagination,
        )

        logger.info(
            f"Writing in destination {self.destination_id} from source iteration {self.buffer.from_iteration} to {self.buffer.to_iteration}"
        )

        success, error_msg = self.write_records(df_destination_records=self.buffer.df_destination_records)

        if success:
            # We wrote records to destination so we keep it
            destination_iteration.records_written = self.buffer.df_destination_records.height
            logger.info(
                f"Successfully wrote {destination_iteration.records_written} records to destination {self.destination_id}"
            )

        else:
            # We failed to write records to destination so we keep the error message
            destination_iteration.error_message = error_msg

        destination_iteration.success = success
        destination_iteration.from_source_iteration = self.buffer.from_iteration
        destination_iteration.to_source_iteration = self.buffer.to_iteration
        destination_iteration.pagination = self.buffer.pagination

        # Update destination cursor
        self.create_cursors(destination_iteration=destination_iteration)

        return destination_iteration

    def write_or_buffer_records(
        self,
        df_destination_records: pl.DataFrame,
        iteration: int,
        last_iteration: bool = False,
        session=None,
        pagination: dict = None,
    ) -> DestinationBufferStatus:
        """Write records to destination or buffer them for the given iteration"""

        # Last iteration, write all records to destination
        if last_iteration:

            if self.buffer.df_destination_records.height == 0 and self.buffer.is_empty:
                logger.info("No records to write to destination, already written, buffer is empty.")
                return DestinationBufferStatus.RECORDS_WRITTEN

            logger.debug("Writing last iteration records to destination")
            assert df_destination_records.height == 0, "Last iteration should not have any records"
            destination_iteration = self.buffer_flush_handler(session=session)

            if destination_iteration.success:
                # Update job status to success

                # Set job status to SUCCEEDED
                # /!\ If the sync mode is STREAM, the job status is kept to RUNNING
                job_status = (
                    JobStatus.SUCCEEDED if self.sync_metadata.sync_mode != SourceSyncModes.STREAM else JobStatus.RUNNING
                )

                self.backend.update_stream_job_status(
                    job_id=self.sync_metadata.job_id,
                    job_status=job_status,
                    session=session,
                )

            # Flush buffer
            self.buffer.flush()
            # Call the finalizing operations to wrap up sync
            self.finalize()
            return DestinationBufferStatus.RECORDS_WRITTEN

        # Don't write empty records to destination
        if df_destination_records.height == 0 and not last_iteration:
            logger.info("No records to write to destination. Check source and queue provider.")
            return DestinationBufferStatus.NO_RECORDS

        # Write records to destination if buffer size is 0 or streaming
        if self.buffer.buffer_size == 0:
            logger.info(f"Writing records to destination {self.destination_id}.")
            self.buffer.add_source_iteration_records_to_buffer(
                iteration=iteration, df_destination_records=df_destination_records, pagination=pagination
            )
            self.buffer_flush_handler(session=session)
            self.buffer.flush()
            return DestinationBufferStatus.RECORDS_WRITTEN

        logger.info(f"Buffer free space {self.buffer.buffer_free_space_pct}%")
        logger.info(f"Buffer current size {round(self.buffer.current_size / 1024 / 1024, 2)}  Mb")
        logger.info(
            f"Buffer ripeness {round(self.buffer.ripeness / 60, 2)} min. Max ripeness {round(self.buffer.buffer_flush_timeout / 60, 2)} min."  # noqa
        )

        # Write buffer to destination if buffer is ripe and create a new buffer for the new iteration
        if self.buffer.is_ripe:
            logger.info(
                f"Buffer is ripe (buffering for longer than buffer_flush_timeout: {self.buffer.buffer_flush_timeout} seconds), writing buffer to destination"  # noqa
            )
            self.buffer_flush_handler(session=session)
            self.buffer.flush()
            self.buffer.add_source_iteration_records_to_buffer(
                iteration=iteration, df_destination_records=df_destination_records, pagination=pagination
            )
            return DestinationBufferStatus.RECORDS_WRITTEN_THEN_BUFFERED

        # Buffer can hold all records from this iteration
        elif self.buffer.buffer_free_space >= df_destination_records.estimated_size(unit="b"):
            self.buffer.add_source_iteration_records_to_buffer(
                iteration=iteration, df_destination_records=df_destination_records, pagination=pagination
            )
            return DestinationBufferStatus.RECORDS_BUFFERED

        # Buffer can contain some records from this iteration
        # For now we will write all records to destination and then buffer the remaining records
        else:
            self.buffer_flush_handler(session=session)
            self.buffer.flush()
            self.buffer.add_source_iteration_records_to_buffer(
                iteration=iteration, df_destination_records=df_destination_records, pagination=pagination
            )
            return DestinationBufferStatus.RECORDS_WRITTEN_THEN_BUFFERED

    def create_cursors(self, destination_iteration: DestinationIteration):
        self.backend.create_destination_cursor(
            job_id=self.sync_metadata.job_id,
            name=self.sync_metadata.name,
            source_name=self.sync_metadata.source_name,
            stream_name=self.sync_metadata.stream_name,
            destination_name=self.sync_metadata.destination_name,
            from_source_iteration=destination_iteration.from_source_iteration,
            to_source_iteration=destination_iteration.to_source_iteration,
            rows_written=destination_iteration.records_written,
            pagination=destination_iteration.pagination,
            success=destination_iteration.success,
        )

    def write_records_and_update_cursor(
        self,
        df_source_records: pl.DataFrame,
        extracted_at: datetime,
        iteration: int,
        last_iteration: bool = False,
        pagination: dict = None,
    ) -> bool:
        """
        Write records to destination and update the cursor for the given iteration.
        Stores the source pagination for recovery purposes.
        """

        # Case when producer failed to fetch data from first iteration
        if iteration == 0 and df_source_records.height == 0:
            logger.warning("Source failed to fetch data from the first iteration, no records will be written.")
            return False

        # Convert to df_destinaton_records
        df_source_records = transform_to_df_destination_records(
            df_source_records=df_source_records, extracted_at=extracted_at
        )

        # Buffer records otherwise write to destination
        self.write_or_buffer_records(
            df_destination_records=df_source_records,
            iteration=iteration,
            last_iteration=last_iteration,
            pagination=pagination,
        )

        return True


class DestinationFactory:
    @staticmethod
    def get_destination(
        sync_metadata: SyncMetadata,
        config: AbstractDestinationConfig,
        backend: AbstractBackend,
        source_callback: AbstractSourceCallback,
    ) -> AbstractDestination:

        if config.name == DestinationTypes.LOGGER:
            from bizon.connectors.destinations.logger.src.destination import (
                LoggerDestination,
            )

            return LoggerDestination(
                sync_metadata=sync_metadata, config=config.config, backend=backend, source_callback=source_callback
            )

        elif config.name == DestinationTypes.BIGQUERY:
            from bizon.connectors.destinations.bigquery.src.destination import (
                BigQueryDestination,
            )

            return BigQueryDestination(
                sync_metadata=sync_metadata, config=config.config, backend=backend, source_callback=source_callback
            )

        elif config.name == DestinationTypes.BIGQUERY_STREAMING:
            from bizon.connectors.destinations.bigquery_streaming.src.destination import (
                BigQueryStreamingDestination,
            )

            return BigQueryStreamingDestination(
                sync_metadata=sync_metadata, config=config.config, backend=backend, source_callback=source_callback
            )

        elif config.name == DestinationTypes.BIGQUERY_STREAMING_V2:
            from bizon.connectors.destinations.bigquery_streaming_v2.src.destination import (
                BigQueryStreamingV2Destination,
            )

            return BigQueryStreamingV2Destination(
                sync_metadata=sync_metadata, config=config.config, backend=backend, source_callback=source_callback
            )

        elif config.name == DestinationTypes.FILE:
            from bizon.connectors.destinations.file.src.destination import (
                FileDestination,
            )

            return FileDestination(
                sync_metadata=sync_metadata, config=config.config, backend=backend, source_callback=source_callback
            )

        raise ValueError(f"Destination {config.name}" f"with params {config} not found")



import json
from enum import Enum
from typing import Any, Generator, List, Optional

from loguru import logger
from pydantic import BaseModel, Field

from bizon.source.config import SourceConfig
from bizon.source.models import SourceIteration, SourceRecord

from .hubspot_base import HubSpotBaseSource
from .models.hs_object import AllObjectProperties, HubSpotProperty

URL_BASE = "https://api.hubapi.com"
URL_GRANTED_SCOPES = f"{URL_BASE}/oauth/v1/access-tokens"
URL_TOKEN_REFRESH = f"{URL_BASE}/oauth/v1/token"


class PropertiesStrategy(str, Enum):
    ALL = "all"
    SELECTED = "selected"


class PropertiesConfig(BaseModel):
    strategy: PropertiesStrategy = Field(PropertiesStrategy.ALL, description="Properties strategy")
    selected_properties: Optional[List[str]] = Field([], description="List of selected properties")


class HubSpotSourceConfig(SourceConfig):
    properties: PropertiesConfig = PropertiesConfig(strategy=PropertiesStrategy.ALL, selected_properties=None)


class HubSpotObjectsSource(HubSpotBaseSource):

    api_version = "v3"

    object_path = f"crm/{api_version}/objects"
    properties_path = f"crm/{api_version}/properties"

    def __init__(self, config: HubSpotSourceConfig):
        super().__init__(config)
        self.config: HubSpotSourceConfig = config
        self.object = self.config.stream
        self.selected_properties = []  # Initialize properties to empty list

        # If we are initializing the pipeline, we retrieve the selected properties from HubSpot
        if config.init_pipeline:
            self.selected_properties = self.get_selected_properties()

    @staticmethod
    def streams() -> List[str]:
        return ["contacts", "companies", "deals"]

    @staticmethod
    def get_config_class() -> SourceConfig:
        return HubSpotSourceConfig

    @property
    def url_list(self) -> str:
        return f"{URL_BASE}/{self.object_path}/{self.object}"

    @property
    def url_list_properties(self) -> str:
        return f"{URL_BASE}/{self.properties_path}/{self.object}"

    @property
    def url_search(self) -> str:
        return f"{URL_BASE}/{self.object_path}/{self.object}/search"

    def _request_api(
        self,
        method: str,
        url: str,
        params: Optional[dict] = None,
        payload: Optional[dict] = None,
        headers=None,
    ) -> Generator[dict, None, None]:

        # Call HubSpot API
        response = self.session.call(
            method=method,
            url=url,
            params=params,
            data=json.dumps(payload),
            headers=headers,
        )
        return response.json()

    def get_selected_properties(self) -> List[str]:
        all_properties = self.list_properties()
        assert (
            len(all_properties.properties) > 0
        ), "No properties found in HubSpot. Which is likely an error with the API."
        properties = []

        if self.config.properties.strategy == "all":
            properties = all_properties.property_names()

        if self.config.properties.strategy == "selected":
            # We check that all properties slected are present in the list of all properties
            for prop in self.config.properties.selected_properties:
                assert prop in all_properties.property_names(), f"Property {prop} is not present in HubSpot."
            properties = self.config.properties.selected_properties

        assert len(properties) > 0, "No properties selected to sync"
        logger.info(f"{len(properties)} selected properties for sync.")
        return properties

    def _get(self, after: str = None) -> Optional[dict]:
        params = {
            "limit": 100,
            "properties": ",".join(self.selected_properties),
        }
        if after:
            params["after"] = after

        return self._request_api(method="GET", url=self.url_list, params=params)

    def get(
        self,
        pagination: dict = None,
    ) -> SourceIteration:
        """Return the next page of data from HubSpot
        Returns:
            dict, Optional[List[dict]]]: Next pagination dict and data
        """

        if not pagination:
            response = self._get()
        else:
            response = self._get(after=pagination["after"])

        return self.parse_response(response)

    def parse_response(self, response: dict) -> SourceIteration:
        # If no response or no results, we return empty dict
        if not response or len(response.get("results", [])) == 0:
            return dict(), []

        # If no next page, we set is_finished to True
        if response.get("paging", dict()).get("next", None) is None:
            return dict(), response["results"]

        # If there is a next page, we set the paging object
        next_pagination_dict = {
            "link": response["paging"]["next"]["link"],
            "after": response["paging"]["next"]["after"],
        }
        return SourceIteration(
            next_pagination=next_pagination_dict,
            records=[
                SourceRecord(
                    id=record["id"],
                    data=record,
                )
                for record in response["results"]
            ],
        )

    def get_total_records_count(self) -> Optional[int]:
        search_response = self._request_api(
            method="POST",
            url=self.url_search,
            payload={"filterGroups": [{"filters": [{"operator": "HAS_PROPERTY", "propertyName": "hs_object_id"}]}]},
        )
        total = search_response["total"]
        logger.info(f"Number of {self.object} in HubSpot: {'{:,}'.format(total).replace(',', ' ')}")
        return total

    def list_properties(self) -> AllObjectProperties:
        response = self._request_api(method="GET", url=self.url_list_properties)
        proprties = response["results"]
        return AllObjectProperties(
            properties=[
                HubSpotProperty(
                    name=hs_property["name"],
                    label=hs_property["label"],
                    field_type=hs_property["fieldType"],
                    type=hs_property["type"],
                    description=hs_property.get("description"),
                )
                for hs_property in proprties
            ]
        )



from typing import List, Literal

from pydantic import BaseModel, Field

from bizon.engine.queue.config import (
    AbastractQueueConfigDetails,
    AbstractQueueConfig,
    QueueTypes,
)


class KafkaQueueConfig(BaseModel):
    bootstrap_server: str = Field(..., description="Kafka bootstrap servers")
    topic: str = Field("bizon", description="Kafka topic")


class KafkaConsumerConfig(BaseModel):
    group_id: str = Field("bizon", description="Kafka group id")
    auto_offset_reset: str = Field("earliest", description="Kafka auto offset reset")
    enable_auto_commit: bool = Field(True, description="Kafka enable auto commit")
    consumer_timeout_ms: int = Field(1000, description="Kafka consumer timeout in milliseconds")


class KafkaConfigDetails(AbastractQueueConfigDetails):
    queue: KafkaQueueConfig = Field(..., description="Kafka queue configuration")
    consumer: KafkaConsumerConfig = Field(KafkaConsumerConfig(), description="Kafka consumer configuration")


class KafkaConfig(AbstractQueueConfig):
    type: Literal[QueueTypes.KAFKA]
    config: KafkaConfigDetails



import json
from datetime import datetime
from typing import Optional, Union

from loguru import logger
from pytz import UTC
from sqlalchemy import Result, Select, create_engine, func, inspect, select, update
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from bizon.engine.backend.backend import AbstractBackend
from bizon.engine.backend.config import BackendTypes
from bizon.engine.backend.models import (
    TABLE_DESTINATION_CURSOR,
    TABLE_SOURCE_CURSOR,
    TABLE_STREAM_INFO,
    Base,
    CursorStatus,
    DestinationCursor,
    JobStatus,
    SourceCursor,
    StreamJob,
)

from .config import BigQueryConfigDetails, PostgresConfigDetails, SQLiteConfigDetails


class SQLAlchemyBackend(AbstractBackend):

    def __init__(self, config: Union[PostgresConfigDetails, SQLiteConfigDetails], type: BackendTypes, **kwargs):
        super().__init__(config, type)

        self._engine = None

        self.config: Union[
            PostgresConfigDetails,
            SQLiteConfigDetails,
            BigQueryConfigDetails,
        ] = config

        self.kwargs = kwargs
        self.session = self.get_session()

    def get_session(self) -> Session:
        """yields a SQLAlchemy connection"""
        engine = self.get_engine()
        session_ = scoped_session(
            sessionmaker(
                bind=engine,
                expire_on_commit=False,
            )
        )

        return session_

    def _get_engine_bigquery(self) -> Engine:
        # If service account key is provided, use it
        if hasattr(self.config, "service_account_key") and self.config.service_account_key:
            return create_engine(
                f"bigquery://{self.config.database}/{self.config.schema_name}",
                echo=self.config.echoEngine,
                credentials_info=self.config.service_account_key,
            )
        # Otherwise we will rely on the default Google Authentication mechanism (e.g. GOOGLE_APPLICATION_CREDENTIALS)
        return create_engine(
            f"bigquery://{self.config.database}/{self.config.schema_name}", echo=self.config.echoEngine
        )

    def _get_engine_postgres(self) -> Engine:
        return create_engine(
            f"postgresql+psycopg2://{self.config.username}:{self.config.password}@{self.config.host}:{self.config.port}/{self.config.database}",
            echo=self.config.echoEngine,
        )

    def _get_engine_sqlite(self) -> Engine:
        return create_engine(
            f"sqlite:///{self.config.database}.sqlite3",
            connect_args={"check_same_thread": False, "timeout": 30},
            pool_size=5,  # Adjust based on expected concurrency
            pool_pre_ping=True,  # Ensures connections are alive
        )

    def _get_engine(self) -> Engine:

        if self.type == BackendTypes.BIGQUERY:
            return self._get_engine_bigquery()

        # Postgres
        if self.type == BackendTypes.POSTGRES:
            return self._get_engine_postgres()

        # SQLite in a file, ok for small tests
        if self.type == BackendTypes.SQLITE:
            return self._get_engine_sqlite()

        # ONLY FOR UNIT TESTS: SQLite in memory
        if self.type == BackendTypes.SQLITE_IN_MEMORY:
            return create_engine(
                f"sqlite:///:memory:",
                echo=self.config.echoEngine,
                connect_args={"check_same_thread": False},
            )

        raise Exception(f"Unsupported database type {self.type}")

    def _check_schema_exist(self):
        if self.type in [BackendTypes.SQLITE, BackendTypes.SQLITE_IN_MEMORY]:
            logger.warning("SQLite does not support schemas")
            return True

        engine = self.get_engine()

        with engine.connect() as connection:
            if not inspect(connection).has_schema(self.config.schema_name):
                logger.error(
                    f"Schema or dataset {self.config.schema_name} does not exist in the database, you need to create it first."
                )
                raise Exception(
                    f"Schema or dataset {self.config.schema_name} does not exist in the database, you need to create it first."
                )

    def get_engine(self) -> Engine:
        """Return the SQLAlchemy engine"""
        if not self._engine:
            self._engine = self._get_engine()

        return self._engine

    #### INIT DATABASE ####

    def create_all_tables(self):
        engine = self.get_engine()
        Base.metadata.create_all(engine)

    def drop_all_tables(self):
        engine = self.get_engine()
        Base.metadata.drop_all(engine)

    def check_prerequisites(self) -> bool:
        """Check if the database contains the necessary tables, return True if entities are present
        Return False if entities are not present, they will be created
        """

        # Check if schema exists
        if self.type != BackendTypes.SQLITE:
            self._check_schema_exist()

        all_entities_exist = True

        engine = self.get_engine()

        # Check if TABLE_STREAM_INFO exists, otherwise create it
        if not inspect(engine).has_table(TABLE_STREAM_INFO):
            all_entities_exist = False
            logger.info(f"Table {TABLE_STREAM_INFO} does not exist in the database, we will create it")

        if not inspect(engine).has_table(TABLE_SOURCE_CURSOR):
            all_entities_exist = False
            logger.info(f"Table {TABLE_SOURCE_CURSOR} does not exist in the database, we will create it")

        if not inspect(engine).has_table(TABLE_DESTINATION_CURSOR):
            all_entities_exist = False
            logger.info(f"Table {TABLE_DESTINATION_CURSOR} does not exist in the database, we will create it")

        return all_entities_exist

    def _add_and_commit(self, obj, session: Optional[Session] = None):
        """Add the object to the session and commit it, return its ID"""
        session = session or self.session
        session.add(obj)
        session.commit()
        return obj

    def _execute(self, select: Select, session: Optional[Session] = None) -> Result:
        session = session or self.session
        result = session.execute(select)
        session.commit()
        return result

    #### STREAM JOB ####

    def create_stream_job(
        self,
        name: str,
        source_name: str,
        stream_name: str,
        sync_mode: str,
        total_records_to_fetch: Optional[int] = None,
        job_status: JobStatus = JobStatus.STARTED,
        session: Optional[Session] = None,
    ) -> StreamJob:
        """Create new StreamJob record in dbt and return its ID"""

        new_stream_job = StreamJob(
            name=name,
            source_name=source_name,
            stream_name=stream_name,
            sync_mode=sync_mode,
            total_records_to_fetch=total_records_to_fetch,
            status=job_status,
        )
        new_stream_job = self._add_and_commit(new_stream_job, session=session)
        logger.debug(f"New streamJob has been created: {new_stream_job}")
        return new_stream_job

    def update_stream_job_status(
        self, job_id: str, job_status: JobStatus, error_message: Optional[str] = None, session: Optional[Session] = None
    ):
        """Update the status of the stream job with the given id"""
        stmt = (
            update(StreamJob)
            .where(StreamJob.id == job_id)
            .values(status=job_status, error_message=error_message, updated_at=datetime.now(tz=UTC))
            .execution_options(synchronize_session="fetch")
        )
        self._execute(stmt, session=session)

    def get_stream_job_by_id(self, job_id: str, session: Optional[Session] = None) -> Optional[StreamJob]:
        """Get the job by its ID"""

        smt = select(StreamJob).filter(
            StreamJob.id == job_id,
        )

        results = self._execute(smt, session=session).one_or_none()

        if results:
            return results[0]
        logger.warning(f"No job found for id={job_id}")
        return None

    def get_running_stream_job(
        self, name: str, source_name: str, stream_name: str, session: Optional[Session] = None
    ) -> Optional[StreamJob]:
        """Get the StreamJob for the given source and stream name"""

        query = select(StreamJob).filter(
            StreamJob.name == name,
            StreamJob.source_name == source_name,
            StreamJob.stream_name == stream_name,
            StreamJob.status == JobStatus.RUNNING,
        )

        job = self._execute(query, session=session).scalar_one_or_none()

        if job:
            return job

        logger.info(f"No running job found for source={source_name} stream={stream_name}")
        return None

    def get_last_successful_stream_job(self, name: str, source_name: str, stream_name: str, session=None):
        """Get the job id for the given source and stream name"""

        query = (
            select(StreamJob)
            .filter(
                StreamJob.name == name,
                StreamJob.source_name == source_name,
                StreamJob.stream_name == stream_name,
                StreamJob.status == JobStatus.SUCCEEDED,
            )
            .order_by(StreamJob.created_at.desc())
            .limit(1)
        )

        job = self._execute(query, session=session).scalar_one_or_none()

        if job:
            return job

        logger.info(f"No last successful job found for source={source_name} stream={stream_name}")
        return None

    #### SOURCE CURSOR ####

    def create_source_cursor(
        self,
        job_id: str,
        name: str,
        source_name: str,
        stream_name: str,
        iteration: int,
        rows_fetched: int,
        next_pagination: dict,
        cursor_status: CursorStatus = CursorStatus.STARTED,
        error_message: Optional[str] = None,
        session: Optional[Session] = None,
    ) -> str:
        """Create a new SourceCursor record in db and return its ID"""
        new_source_cursor = SourceCursor(
            job_id=job_id,
            name=name,
            source_name=source_name,
            stream_name=stream_name,
            iteration=iteration,
            rows_fetched=rows_fetched,
            next_pagination=json.dumps(next_pagination),
            status=cursor_status,
            error_message=error_message,
        )
        new_source_cursor_id = self._add_and_commit(new_source_cursor, session=session).id
        logger.debug(f"New streamCursor has been created with id={new_source_cursor_id}")
        return new_source_cursor_id

    def update_source_cursor_status(
        self,
        cursor_id: str,
        cursor_status: CursorStatus,
        error_message: Optional[str] = None,
        session: Optional[Session] = None,
    ):
        """Update the status of the stream cursor with the given id"""
        stmt = (
            update(SourceCursor)
            .where(SourceCursor.id == cursor_id)
            .values(status=cursor_status, error_message=error_message, updated_at=datetime.now(tz=UTC))
            .execution_options(synchronize_session="fetch")
        )
        self._execute(stmt, session=session)

    def get_source_cursor_by_id(self, cursor_id: str, session: Optional[Session] = None) -> Optional[SourceCursor]:
        """Get the cursor by its ID"""

        smt = select(SourceCursor).filter(
            SourceCursor.id == cursor_id,
        )

        results = self._execute(smt, session=session).one_or_none()
        if results:
            return results[0]
        logger.warning(f"No cursor found for id={cursor_id}")
        return None

    def get_last_cursor_by_job_id(self, job_id: str, session: Optional[Session] = None) -> Optional[DestinationCursor]:
        """Get the last cursor we should start source from for the given job id"""

        smt = (
            select(DestinationCursor)
            .filter(
                DestinationCursor.job_id == job_id,
                DestinationCursor.success == True,  # noqa
            )
            .order_by(DestinationCursor.to_source_iteration.desc())
        )

        results = self._execute(smt, session=session).first()
        if results:
            return results[0]

        logger.warning(f"No last cursor found for job_id={job_id}")
        return None

    def get_last_source_cursor_by_stream_and_iteration(
        self, name: str, source_name: str, stream_name: str, iteration: int, session: Optional[Session] = None
    ) -> Optional[SourceCursor]:
        """Get the last cursor for the given stream and iteration"""

        smt = (
            select(SourceCursor)
            .filter(
                SourceCursor.name == name,
                SourceCursor.source_name == source_name,
                SourceCursor.stream_name == stream_name,
                SourceCursor.iteration == iteration,
            )
            .order_by(SourceCursor.created_at.desc())
        )

        results = self._execute(smt, session=session).first()
        if results:
            return results[0]

        logger.warning(f"No last cursor found for source={source_name} stream={stream_name} iteration={iteration}")
        return None

    def create_destination_cursor(
        self,
        job_id: str,
        name: str,
        source_name: str,
        stream_name: str,
        destination_name: str,
        from_source_iteration: int,
        to_source_iteration: int,
        rows_written: int,
        success: bool,
        pagination: Optional[dict] = None,
        session: Session | None = None,
    ) -> DestinationCursor:

        destination_cursor = DestinationCursor(
            name=name,
            source_name=source_name,
            stream_name=stream_name,
            destination_name=destination_name,
            job_id=job_id,
            from_source_iteration=from_source_iteration,
            to_source_iteration=to_source_iteration,
            rows_written=rows_written,
            pagination=json.dumps(pagination) if pagination else None,
            success=success,
        )
        new_destination_cursor = self._add_and_commit(destination_cursor, session=session)
        logger.debug(f"New Destination Cursor has been created: {new_destination_cursor}")
        return new_destination_cursor

    def get_destination_cursor_by_id(
        self, cursor_id: str, session: Optional[Session] = None
    ) -> Optional[DestinationCursor]:
        """Get the destination cursor by its ID"""
        smt = select(DestinationCursor).filter(
            DestinationCursor.id == cursor_id,
        )
        results = self._execute(smt, session=session).one_or_none()

        if results:
            return results[0]
        logger.warning(f"No job found for id={cursor_id}")
        return None

    def get_number_of_written_rows_for_job(self, job_id: str, session: Optional[Session] = None) -> Optional[int]:
        """Get the number of written rows for the given job"""
        smt = select(func.sum(DestinationCursor.rows_written)).filter(
            DestinationCursor.job_id == job_id,
            DestinationCursor.success == True,  # noqa
        )
        results = self._execute(smt, session=session).one_or_none()
        if results:
            return results[0]
        else:
            raise Exception(f"No rows written found for job_id={job_id}")



from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field

from bizon.engine.pipeline.models import PipelineReturnStatus


class RunnerTypes(str, Enum):
    THREAD = "thread"
    PROCESS = "process"
    STREAM = "stream"


class LoggerLevel(str, Enum):
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class RunnerFuturesConfig(BaseModel):
    max_workers: Optional[int] = Field(
        description="Number of workers to use for the runner",
        default=2,
    )
    consumer_start_delay: Optional[int] = Field(
        description="Duration in seconds to wait before starting the consumer thread",
        default=2,
    )
    is_alive_check_interval: Optional[int] = Field(
        description="Duration in seconds to wait between checking if the producer and consumer threads are still running",
        default=2,
    )


class RunnerConfig(BaseModel):

    type: RunnerTypes = Field(
        description="Runner to use for the pipeline",
        default=RunnerTypes.THREAD,
    )

    config: Optional[RunnerFuturesConfig] = Field(
        description="Runner configuration",
        default=RunnerFuturesConfig(),
    )

    log_level: LoggerLevel = Field(
        description="Logging level",
        default=LoggerLevel.INFO,
    )


class RunnerStatus(BaseModel):
    producer: Optional[PipelineReturnStatus] = None
    consumer: Optional[PipelineReturnStatus] = None
    stream: Optional[PipelineReturnStatus] = None

    def __init__(self, **data):
        super().__init__(**data)
        if not ((self.producer is not None and self.consumer is not None) or self.stream is not None):
            raise ValueError("Either both producer and consumer must be set, or stream must be set")

    @property
    def is_success(self):
        if self.stream is not None:
            return self.stream == PipelineReturnStatus.SUCCESS
        return self.producer == PipelineReturnStatus.SUCCESS and self.consumer == PipelineReturnStatus.SUCCESS

    def to_string(self):
        if self.stream is not None:
            return f"Pipeline finished with status {'Success' if self.is_success else 'Failure'} (Stream: {self.stream.value})"
        return (
            f"Pipeline finished with status {'Success' if self.is_success else 'Failure'} "
            f"(Producer: {self.producer.value}, Consumer: {self.consumer.value})"
        )



from typing import Literal

from pydantic import Field

from bizon.engine.backend.config import (
    AbstractBackendConfig,
    AbstractBackendConfigDetails,
    BackendTypes,
)


class SQLAlchemyConfigDetails(AbstractBackendConfigDetails):
    echoEngine: bool = Field(False, description="Echo the engine in logs")


## POSTGRES ##
class PostgresConfigDetails(SQLAlchemyConfigDetails):
    host: str = Field(
        description="Host of the database",
        default=...,
    )
    port: int = Field(
        description="Port of the database",
        default=...,
    )
    username: str = Field(
        description="Username to connect to the database",
        default=...,
    )
    password: str = Field(
        description="Password to connect to the database",
        default=...,
    )


class PostgresSQLAlchemyConfig(AbstractBackendConfig):
    type: Literal[BackendTypes.POSTGRES]
    config: PostgresConfigDetails


## SQLITE ##
class SQLiteConfigDetails(SQLAlchemyConfigDetails):
    pass


class SQLiteSQLAlchemyConfig(AbstractBackendConfig):
    type: Literal[BackendTypes.SQLITE]
    config: SQLiteConfigDetails


class SQLiteInMemoryConfig(AbstractBackendConfig):
    type: Literal[BackendTypes.SQLITE_IN_MEMORY]
    config: SQLiteConfigDetails


## BIGQUERY ##
class BigQueryConfigDetails(SQLAlchemyConfigDetails):

    database: str = Field(
        description="GCP Project name",
        default=...,
    )

    schema_name: str = Field(
        description="BigQuery Dataset name",
        default=...,
        alias="schema",
    )

    service_account_key: str = Field(
        description="Service Account Key JSON string. If empty it will be infered",
        default="",
    )


class BigQuerySQLAlchemyConfig(AbstractBackendConfig):
    type: Literal[BackendTypes.BIGQUERY]
    config: SQLAlchemyConfigDetails



import json
import re
from typing import Any, Counter, List, Tuple, Type
from uuid import uuid4

import google.auth
import gspread
import gspread.utils
from google.oauth2.service_account import Credentials
from loguru import logger
from pydantic import Field
from requests.auth import AuthBase

from bizon.source.config import SourceConfig
from bizon.source.models import SourceIteration, SourceRecord
from bizon.source.source import AbstractSource


class GsheetsSourceConfig(SourceConfig):
    worksheet_name: str = Field(description="Name of the worksheet to fetch data from", default=...)
    spreadsheet_url: str = Field(description="URL of the spreadsheet", default=...)
    service_account_key: str = Field(
        description="Service Account Key JSON string. If empty it will be infered",
        default="",
    )
    column_names: list[str] = Field(
        description="Column names to fetch from the worksheet, if empty all columns will be fetched",
        default=[],
    )
    convert_column_names_to_sql_format: bool = Field(
        description="Convert column names to SQL format (lowercase, no spaces, etc)",
        default=True,
    )


class GsheetsSource(AbstractSource):

    def __init__(self, config: GsheetsSourceConfig):
        super().__init__(config)
        self.config: GsheetsSourceConfig = config
        self.normalization_pattern = re.compile(r"(?<!^)(?=[A-Z])")

    @staticmethod
    def streams() -> List[str]:
        return ["worksheet"]

    @staticmethod
    def get_config_class() -> SourceConfig:
        return GsheetsSourceConfig

    def get_gspread_client(self) -> gspread.client.Client:

        if self.config.service_account_key:
            # use creds to create a client to interact with the Google Drive API
            credentials_dict = json.loads(self.config.service_account_key)
            credentials = Credentials.from_service_account_info(credentials_dict)
            credentials = credentials.with_scopes(gspread.auth.READONLY_SCOPES)
            gc = gspread.authorize(credentials)
        else:
            # use default credentials
            credentials, project_id = google.auth.default(scopes=gspread.auth.READONLY_SCOPES)
            gc = gspread.authorize(credentials)
        return gc

    def check_connection(self) -> Tuple[bool | Any | None]:
        gc = self.get_gspread_client()

        # Open a sheet from a spreadsheet in one go
        sh = gc.open_by_url(self.config.spreadsheet_url)
        try:
            _ = sh.worksheet(self.config.worksheet_name)
        except gspread.WorksheetNotFound:
            return False, f"Worksheet not found, available worksheets: {sh.worksheets()}"
        return True, None

    def get_authenticator(self) -> AuthBase | None:
        return None

    def get_total_records_count(self) -> int | None:
        return None

    def check_column_names_are_unique(self, column_names: List[str]) -> bool:
        """Check if all column names are unique, otherwise raise an error listing duplicates"""
        if len(column_names) != len(set(column_names)):
            duplicates = [item for item, count in Counter(column_names).items() if count > 1]
            logger.error(
                f"Column names are not unique: {duplicates}, found following columns: {column_names}."
                f"Please provide unique column names in the config using `column_names` param or remove duplicated column in sheets."
            )
            raise ValueError(f"Column names are not unique: {duplicates}, found following columns: {column_names}.")
        return True

    def normalize_record_to_sql_format_inplace(self, records: List[dict]) -> dict:
        """Normalize record to SQL format inplace"""
        for i, record in enumerate(records):
            records[i] = {
                self.normalization_pattern.sub("_", k).lower().replace(" ", "_"): v for k, v in record.items()
            }

    def keep_only_selected_columns(self, records: List[dict], column_names: List[str]) -> List[dict]:
        """Keep only selected columns in records"""
        return [{k: v for k, v in record.items() if k in column_names} for record in records]

    def get(self, pagination: dict = None) -> SourceIteration:
        gc = self.get_gspread_client()
        worksheet = gc.open_by_url(self.config.spreadsheet_url).worksheet(self.config.worksheet_name)

        worksheet_column_names = worksheet.row_values(1)

        if self.config.column_names:
            logger.info(f"Using provided column names: {','.join(self.config.column_names)}")

            self.check_column_names_are_unique(self.config.column_names)

            # Ensure column names are all present in worksheet_column_names
            for column_name in self.config.column_names:
                if column_name not in worksheet_column_names:
                    raise ValueError(f"Column name {column_name} not found in worksheet")

            # Get all records
            all_records = worksheet.get_all_records(expected_headers=self.config.column_names)
            all_records = self.keep_only_selected_columns(all_records, self.config.column_names)

        else:
            # Ensure column names are unique
            self.check_column_names_are_unique(worksheet_column_names)

            # Get all records
            all_records = worksheet.get_all_records()

        if self.config.convert_column_names_to_sql_format:
            self.normalize_record_to_sql_format_inplace(all_records)

        return SourceIteration(
            records=[SourceRecord(id=uuid4().hex, data=record) for record in all_records],
            next_pagination={},
        )



from typing import Any, Mapping, Optional

from pydantic import BaseModel, Field
from requests import PreparedRequest
from requests.auth import AuthBase


class CookiesAuthParams(BaseModel):
    cookies: Mapping[str, str] = Field(..., description="Cookies to be attached to the request")
    headers: Optional[Mapping[str, Any]] = Field({}, description="Headers to be attached to the request")


class CookiesAuthenticator(AuthBase):
    def __init__(self, params: CookiesAuthParams):
        super().__init__()
        self.cookies = params.cookies
        self.headers = params.headers

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        request.prepare_cookies(self.cookies)

        if self.headers:
            request.headers.update(self.headers)
        return request



import json
from typing import Union

from kafka import KafkaProducer
from loguru import logger

from bizon.destination.destination import AbstractDestination
from bizon.engine.queue.config import QUEUE_TERMINATION, QueueMessage
from bizon.engine.queue.queue import AbstractQueue

from .config import KafkaConfigDetails
from .consumer import KafkaConsumer_


class KafkaQueue(AbstractQueue):

    def __init__(self, config: KafkaConfigDetails) -> None:
        super().__init__(config)
        self.config: KafkaConfigDetails = config

    def connect(self):
        self.producer = self.get_kafka_producer()

    def get_consumer(self, destination: AbstractDestination) -> KafkaConsumer_:
        return KafkaConsumer_(config=self.config, destination=destination)

    def get_kafka_producer(self) -> KafkaProducer:
        return KafkaProducer(
            bootstrap_servers=self.config.queue.bootstrap_server,
            value_serializer=lambda m: json.dumps(m).encode("utf-8"),
        )

    @staticmethod
    def on_success(metadata):
        logger.info(f"Message produced to topic '{metadata.topic}' at offset {metadata.offset}")

    @staticmethod
    def on_error(e):
        logger.error(f"Error sending message: {e}")

    def get_size(self) -> Union[int, None]:
        return None

    def put_queue_message(self, queue_message: QueueMessage):
        future = self.producer.send(
            topic=self.config.queue.topic,
            value=queue_message.model_dump(),
        )
        future.add_callback(self.on_success)
        future.add_errback(self.on_error)
        self.producer.flush()

    def get(self) -> QueueMessage:
        raise NotImplementedError("Kafka does not support getting messages from here. Use KafkaConsumer instead.")

    def terminate(self, iteration: int) -> bool:
        self.put(source_records=[], iteration=iteration, signal=QUEUE_TERMINATION)
        self.producer.close()
        logger.debug("Terminating Kafka producer ...")
        return True



import os
import tempfile
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import List, Tuple, Type

import polars as pl
from google.api_core.exceptions import NotFound
from google.cloud import bigquery, bigquery_storage_v1
from google.cloud.bigquery import DatasetReference, TimePartitioning
from google.cloud.bigquery_storage_v1.types import (
    AppendRowsRequest,
    ProtoRows,
    ProtoSchema,
)
from google.protobuf.json_format import ParseDict
from google.protobuf.message import Message
from loguru import logger

from bizon.common.models import SyncMetadata
from bizon.destination.destination import AbstractDestination
from bizon.engine.backend.backend import AbstractBackend
from bizon.source.callback import AbstractSourceCallback

from .config import BigQueryStreamingV2ConfigDetails
from .proto_utils import get_proto_schema_and_class


class BigQueryStreamingV2Destination(AbstractDestination):

    # Add constants for limits
    MAX_ROWS_PER_REQUEST = 5000  # 5000 (max is 10000)
    MAX_REQUEST_SIZE_BYTES = 5 * 1024 * 1024  # 5 MB (max is 10MB)
    MAX_ROW_SIZE_BYTES = 0.9 * 1024 * 1024  # 1 MB

    def __init__(
        self,
        sync_metadata: SyncMetadata,
        config: BigQueryStreamingV2ConfigDetails,
        backend: AbstractBackend,
        source_callback: AbstractSourceCallback,
    ):  # type: ignore
        super().__init__(sync_metadata, config, backend, source_callback)
        self.config: BigQueryStreamingV2ConfigDetails = config

        if config.authentication and config.authentication.service_account_key:
            with tempfile.NamedTemporaryFile(delete=False) as temp:
                temp.write(config.authentication.service_account_key.encode())
                temp_file_path = temp.name
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = temp_file_path

        self.project_id = config.project_id
        self.bq_client = bigquery.Client(project=self.project_id)
        self.bq_storage_client = bigquery_storage_v1.BigQueryWriteClient()
        self.dataset_id = config.dataset_id
        self.dataset_location = config.dataset_location
        self.bq_max_rows_per_request = config.bq_max_rows_per_request

    @property
    def table_id(self) -> str:
        tabled_id = f"{self.sync_metadata.source_name}_{self.sync_metadata.stream_name}"
        return self.destination_id or f"{self.project_id}.{self.dataset_id}.{tabled_id}"

    def get_bigquery_schema(self) -> List[bigquery.SchemaField]:

        if self.config.unnest:
            if len(list(self.record_schemas.keys())) == 1:
                self.destination_id = list(self.record_schemas.keys())[0]

            return [
                bigquery.SchemaField(
                    name=col.name,
                    field_type=col.type,
                    mode=col.mode,
                    description=col.description,
                    default_value_expression=col.default_value_expression,
                )
                for col in self.record_schemas[self.destination_id]
            ]

        # Case we don't unnest the data
        else:
            return [
                bigquery.SchemaField("_source_record_id", "STRING", mode="REQUIRED"),
                bigquery.SchemaField("_source_timestamp", "TIMESTAMP", mode="REQUIRED"),
                bigquery.SchemaField("_source_data", "JSON", mode="NULLABLE"),
                bigquery.SchemaField("_bizon_extracted_at", "TIMESTAMP", mode="REQUIRED"),
                bigquery.SchemaField(
                    "_bizon_loaded_at", "TIMESTAMP", mode="REQUIRED", default_value_expression="CURRENT_TIMESTAMP()"
                ),
                bigquery.SchemaField("_bizon_id", "STRING", mode="REQUIRED"),
            ]

    def check_connection(self) -> bool:
        dataset_ref = DatasetReference(self.project_id, self.dataset_id)

        try:
            self.bq_client.get_dataset(dataset_ref)
        except NotFound:
            dataset = bigquery.Dataset(dataset_ref)
            dataset.location = self.dataset_location
            dataset = self.bq_client.create_dataset(dataset)
        return True

    def append_rows_to_stream(
        self,
        write_client: bigquery_storage_v1.BigQueryWriteClient,
        stream_name: str,
        proto_schema: ProtoSchema,
        serialized_rows: List[bytes],
    ):
        request = AppendRowsRequest(
            write_stream=stream_name,
            proto_rows=AppendRowsRequest.ProtoData(
                rows=ProtoRows(serialized_rows=serialized_rows),
                writer_schema=proto_schema,
            ),
        )
        response = write_client.append_rows(iter([request]))
        return response.code().name

    def safe_cast_record_values(self, row: dict):
        for col in self.record_schemas[self.destination_id]:
            if col.type in ["TIMESTAMP", "DATETIME"] and col.default_value_expression is None:
                if isinstance(row[col.name], int):
                    if row[col.name] > datetime(9999, 12, 31).timestamp():
                        row[col.name] = datetime.fromtimestamp(row[col.name] / 1_000_000).strftime(
                            "%Y-%m-%d %H:%M:%S.%f"
                        )
                    else:
                        try:
                            row[col.name] = datetime.fromtimestamp(row[col.name]).strftime("%Y-%m-%d %H:%M:%S.%f")
                        except ValueError:
                            error_message = (
                                f"Error casting timestamp for destination '{self.destination_id}' column '{col.name}'. "
                                f"Invalid timestamp value: {row[col.name]} ({type(row[col.name])}). "
                                "Consider using a transformation."
                            )
                            logger.error(error_message)
                            raise ValueError(error_message)
        return row

    @staticmethod
    def to_protobuf_serialization(TableRowClass: Type[Message], row: dict) -> bytes:
        """Convert a row to a Protobuf serialization."""
        record = ParseDict(row, TableRowClass())
        return record.SerializeToString()

    def load_to_bigquery_via_streaming(self, df_destination_records: pl.DataFrame) -> str:

        # TODO: for now no clustering keys
        clustering_keys = []

        # Create table if it doesnt exist
        schema = self.get_bigquery_schema()
        table = bigquery.Table(self.table_id, schema=schema)
        time_partitioning = TimePartitioning(
            field=self.config.time_partitioning.field, type_=self.config.time_partitioning.type
        )
        table.time_partitioning = time_partitioning

        # Override bigquery client with project's destination id
        if self.destination_id:
            project, dataset, table_name = self.destination_id.split(".")
            self.bq_client = bigquery.Client(project=project)

        table = self.bq_client.create_table(table, exists_ok=True)

        # Create the stream
        if self.destination_id:
            project, dataset, table_name = self.destination_id.split(".")
            write_client = bigquery_storage_v1.BigQueryWriteClient()
            parent = write_client.table_path(project, dataset, table_name)
        else:
            write_client = self.bq_storage_client
            parent = write_client.table_path(self.project_id, self.dataset_id, self.destination_id)

        stream_name = f"{parent}/_default"

        # Generating the protocol buffer representation of the message descriptor.
        proto_schema, TableRow = get_proto_schema_and_class(schema, clustering_keys)

        if self.config.unnest:
            serialized_rows = [
                self.to_protobuf_serialization(TableRowClass=TableRow, row=self.safe_cast_record_values(row))
                for row in df_destination_records["source_data"].str.json_decode(infer_schema_length=None).to_list()
            ]
        else:
            df_destination_records = df_destination_records.with_columns(
                pl.col("bizon_extracted_at").dt.strftime("%Y-%m-%d %H:%M:%S").alias("bizon_extracted_at"),
                pl.col("bizon_loaded_at").dt.strftime("%Y-%m-%d %H:%M:%S").alias("bizon_loaded_at"),
                pl.col("source_timestamp").dt.strftime("%Y-%m-%d %H:%M:%S").alias("source_timestamp"),
            )
            df_destination_records = df_destination_records.rename(
                {
                    "bizon_id": "_bizon_id",
                    "bizon_extracted_at": "_bizon_extracted_at",
                    "bizon_loaded_at": "_bizon_loaded_at",
                    "source_record_id": "_source_record_id",
                    "source_timestamp": "_source_timestamp",
                    "source_data": "_source_data",
                }
            )

            serialized_rows = [
                self.to_protobuf_serialization(TableRowClass=TableRow, row=row)
                for row in df_destination_records.iter_rows(named=True)
            ]

        results = []
        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self.append_rows_to_stream, write_client, stream_name, proto_schema, batch_rows)
                for batch_rows in self.batch(serialized_rows)
            ]
            for future in futures:
                results.append(future.result())

        assert all([r == "OK" for r in results]) is True, "Failed to append rows to stream"

    def write_records(self, df_destination_records: pl.DataFrame) -> Tuple[bool, str]:
        self.load_to_bigquery_via_streaming(df_destination_records=df_destination_records)
        return True, ""

    def batch(self, iterable):
        """
        Yield successive batches respecting both row count and size limits.
        """
        current_batch = []
        current_batch_size = 0
        large_rows = []

        for item in iterable:
            # Estimate the size of the item (as JSON)
            item_size = len(str(item).encode("utf-8"))

            # If adding this item would exceed either limit, yield current batch and start new one
            if (
                len(current_batch) >= self.MAX_ROWS_PER_REQUEST
                or current_batch_size + item_size > self.MAX_REQUEST_SIZE_BYTES
            ):
                logger.debug(f"Yielding batch of {len(current_batch)} rows, size: {current_batch_size/1024/1024:.2f}MB")
                yield {"stream_batch": current_batch, "json_batch": large_rows}
                current_batch = []
                current_batch_size = 0
                large_rows = []

            if item_size > self.MAX_ROW_SIZE_BYTES:
                large_rows.append(item)
                logger.debug(f"Large row detected: {item_size} bytes")
            else:
                current_batch.append(item)
                current_batch_size += item_size

        # Yield the last batch
        if current_batch:
            logger.debug(
                f"Yielding streaming batch of {len(current_batch)} rows, size: {current_batch_size/1024/1024:.2f}MB"
            )
            logger.debug(f"Yielding large rows batch of {len(large_rows)} rows")
            yield {"stream_batch": current_batch, "json_batch": large_rows}



import traceback
from datetime import datetime
from functools import lru_cache
from typing import Any, List, Mapping, Tuple

import orjson
from avro.schema import Schema, parse
from confluent_kafka import (
    Consumer,
    KafkaError,
    KafkaException,
    Message,
    TopicPartition,
)
from loguru import logger
from pydantic import BaseModel
from pytz import UTC

from bizon.source.auth.config import AuthType
from bizon.source.callback import AbstractSourceCallback
from bizon.source.config import SourceSyncModes
from bizon.source.models import SourceIteration, SourceRecord
from bizon.source.source import AbstractSource

from .callback import KafkaSourceCallback
from .config import KafkaSourceConfig, MessageEncoding, SchemaRegistryType
from .decode import (
    Hashabledict,
    decode_avro_message,
    get_header_bytes,
    parse_global_id_from_serialized_message,
)


class SchemaNotFound(Exception):
    pass


class OffsetPartition(BaseModel):
    first: int
    last: int
    to_fetch: int = 0


class TopicOffsets(BaseModel):
    name: str
    partitions: Mapping[int, OffsetPartition]

    def set_partition_offset(self, index: int, offset: int):
        self.partitions[index].to_fetch = offset

    def get_partition_offset(self, index: int) -> int:
        return self.partitions[index].to_fetch

    @property
    def total_offset(self) -> int:
        return sum([partition.last for partition in self.partitions.values()])


class KafkaSource(AbstractSource):

    def __init__(self, config: KafkaSourceConfig):
        super().__init__(config)

        self.config: KafkaSourceConfig = config

        # Kafka consumer configuration.
        if self.config.authentication.type == AuthType.BASIC:
            self.config.consumer_config["sasl.mechanisms"] = "PLAIN"
            self.config.consumer_config["sasl.username"] = self.config.authentication.params.username
            self.config.consumer_config["sasl.password"] = self.config.authentication.params.password

        # Set the bootstrap servers and group id
        self.config.consumer_config["group.id"] = self.config.group_id
        self.config.consumer_config["bootstrap.servers"] = self.config.bootstrap_servers

        # Consumer instance
        self.consumer = Consumer(self.config.consumer_config, logger=logger)

        self.topic_map = {topic.name: topic.destination_id for topic in self.config.topics}

    @staticmethod
    def streams() -> List[str]:
        return ["topic"]

    def get_authenticator(self):
        # We don't use HTTP authentication for Kafka
        # We use confluence_kafka library to authenticate
        pass

    @staticmethod
    def get_config_class() -> AbstractSource:
        return KafkaSourceConfig

    def get_source_callback_instance(self) -> AbstractSourceCallback:
        """Return an instance of the source callback, used to commit the offsets of the iterations"""
        return KafkaSourceCallback(config=self.config)

    def check_connection(self) -> Tuple[bool | Any | None]:
        """Check the connection to the Kafka source"""

        logger.info(f"Found: {len(self.consumer.list_topics().topics)} topics")

        topics = self.consumer.list_topics().topics

        config_topics = [topic.name for topic in self.config.topics]

        for topic in config_topics:
            if topic not in topics:
                logger.error(f"Topic {topic} not found, available topics: {topics.keys()}")
                return False, f"Topic {topic} not found"

            logger.info(f"Topic {topic} has {len(topics[topic].partitions)} partitions")

        return True, None

    def get_number_of_partitions(self, topic: str) -> int:
        """Get the number of partitions for the topic"""
        return len(self.consumer.list_topics().topics[topic].partitions)

    def get_offset_partitions(self, topic: str) -> TopicOffsets:
        """Get the offsets for each partition of the topic"""

        partitions: Mapping[int, OffsetPartition] = {}

        for i in range(self.get_number_of_partitions(topic)):
            offsets = self.consumer.get_watermark_offsets(TopicPartition(topic, i))
            partitions[i] = OffsetPartition(first=offsets[0], last=offsets[1])

        return TopicOffsets(name=topic, partitions=partitions)

    def get_total_records_count(self) -> int | None:
        """Get the total number of records in the topic, sum of offsets for each partition"""
        # Init the consumer
        total_records = 0
        for topic in [topic.name for topic in self.config.topics]:
            total_records += self.get_offset_partitions(topic).total_offset
        return total_records

    @lru_cache(maxsize=None)
    def get_schema_from_registry(self, global_id: int) -> Tuple[Hashabledict, Schema]:
        """Get the schema from the registry, return a hashable dict and an avro schema object"""

        # Apicurio
        if self.config.authentication.schema_registry_type == SchemaRegistryType.APICURIO:
            try:
                response = self.session.get(
                    f"{self.config.authentication.schema_registry_url}/apis/registry/v2/ids/globalIds/{global_id}",
                    auth=(
                        self.config.authentication.schema_registry_username,
                        self.config.authentication.schema_registry_password,
                    ),
                )
                if response.status_code == 404:
                    raise SchemaNotFound(f"Schema with global id {global_id} not found")

                schema_dict = response.json()

            except Exception as e:
                logger.error(traceback.format_exc())
                raise e

            # Add a name field to the schema as needed by fastavro
            schema_dict["name"] = "Envelope"

            # Convert the schema dict to an avro schema object
            avro_schema = parse(orjson.dumps(schema_dict))

            # Convert the schema dict to a hashable dict
            hashable_dict_schema = Hashabledict(schema_dict)

            return hashable_dict_schema, avro_schema

        else:
            raise ValueError(f"Schema registry type {self.config.authentication.schema_registry_type} not supported")

    def decode_avro(self, message: Message) -> dict:
        # Get the header bytes and the global id from the message
        header_message_bytes = get_header_bytes(
            nb_bytes_schema_id=self.config.nb_bytes_schema_id, message=message.value()
        )
        global_id = parse_global_id_from_serialized_message(
            nb_bytes_schema_id=self.config.nb_bytes_schema_id,
            header_message_bytes=header_message_bytes,
        )

        try:
            hashable_dict_schema, avro_schema = self.get_schema_from_registry(global_id=global_id)
        except SchemaNotFound as e:
            logger.error(
                (
                    f"Message on topic {message.topic()} partition {message.partition()} at offset {message.offset()} has a  SchemaID of {global_id} which is not found in Registry."
                    f"message value: {message.value()}."
                )
            )
            logger.error(traceback.format_exc())
            raise e

        return decode_avro_message(
            message=message,
            nb_bytes_schema_id=self.config.nb_bytes_schema_id,
            hashable_dict_schema=hashable_dict_schema,
            avro_schema=avro_schema,
        )

    def decode_utf_8(self, message: Message):
        # Decode the message as utf-8
        return orjson.loads(message.value().decode("utf-8"))

    def decode(self, message):
        if self.config.message_encoding == MessageEncoding.AVRO:
            return self.decode_avro(message)

        elif self.config.message_encoding == MessageEncoding.UTF_8:
            return self.decode_utf_8(message)

        else:
            raise ValueError(f"Message encoding {self.config.message_encoding} not supported")

    def parse_encoded_messages(self, encoded_messages: list) -> List[SourceRecord]:

        records = []

        for message in encoded_messages:

            if message.error():
                # If the message is too large, we skip it and update the offset
                if message.error().code() == KafkaError.MSG_SIZE_TOO_LARGE:
                    logger.warning(
                        (
                            f"Message for topic {message.topic()} partition {message.partition()} and offset {message.offset()} has been skipped. "
                            f"Raised MSG_SIZE_TOO_LARGE, we suppose the message does not exist. Double-check in Confluent Cloud."
                        )
                    )
                    continue

                logger.error(
                    (
                        f"Error while consuming message for topic {message.topic()} partition {message.partition()} and offset {message.offset()}: "
                        f"{message.error()}"
                    )
                )
                raise KafkaException(message.error())

            # We skip tombstone messages
            if self.config.skip_message_empty_value and not message.value():
                logger.debug(
                    f"Message for topic {message.topic()} partition {message.partition()} and offset {message.offset()} is empty, skipping."
                )
                continue

            # Decode the message
            try:

                data = {
                    "topic": message.topic(),
                    "offset": message.offset(),
                    "partition": message.partition(),
                    "timestamp": message.timestamp()[1],
                    "keys": orjson.loads(message.key().decode("utf-8")) if message.key() else {},
                    "headers": (
                        {key: value.decode("utf-8") for key, value in message.headers()} if message.headers() else {}
                    ),
                    "value": self.decode(message),
                }

                records.append(
                    SourceRecord(
                        id=f"partition_{message.partition()}_offset_{message.offset()}",
                        timestamp=datetime.fromtimestamp(message.timestamp()[1] / 1000, tz=UTC),
                        data=data,
                        destination_id=self.topic_map[message.topic()],
                    )
                )

            except Exception as e:
                logger.error(
                    (
                        f"Error while decoding message for topic {message.topic()} on partition {message.partition()}: {e} at offset {message.offset()} "
                        f"with value: {message.value()} and key: {message.key()}"
                    )
                )
                # Try to parse error message from the message
                try:
                    message_raw_text = message.value().decode("utf-8")
                    logger.error(f"Parsed Kafka value: {message_raw_text}")
                except UnicodeDecodeError:
                    logger.error("Message is not a valid UTF-8 string")

                logger.error(traceback.format_exc())
                raise e

        return records

    def read_topics_manually(self, pagination: dict = None) -> SourceIteration:
        """Read the topics manually, we use consumer.assign to assign to the partitions and get the offsets"""

        assert len(self.config.topics) == 1, "Only one topic is supported for manual mode"

        # We will use the first topic for the manual mode
        topic = self.config.topics[0]

        nb_partitions = self.get_number_of_partitions(topic=topic.name)

        # Setup offset_pagination
        self.topic_offsets = (
            TopicOffsets.model_validate(pagination) if pagination else self.get_offset_partitions(topic=topic.name)
        )

        self.consumer.assign(
            [
                TopicPartition(topic.name, partition, self.topic_offsets.get_partition_offset(partition))
                for partition in range(nb_partitions)
            ]
        )

        t1 = datetime.now()
        encoded_messages = self.consumer.consume(self.config.batch_size, timeout=self.config.consumer_timeout)
        logger.info(f"Kafka consumer read : {len(encoded_messages)} messages in {datetime.now() - t1}")

        records = self.parse_encoded_messages(encoded_messages)

        # Update the offset for the partition
        if not records:
            logger.info("No new records found, stopping iteration")
            return SourceIteration(
                next_pagination={},
                records=[],
            )

        # Update the offset for the partition
        self.topic_offsets.set_partition_offset(encoded_messages[-1].partition(), encoded_messages[-1].offset() + 1)

        return SourceIteration(
            next_pagination=self.topic_offsets.model_dump(),
            records=records,
        )

    def read_topics_with_subscribe(self, pagination: dict = None) -> SourceIteration:
        """Read the topics with the subscribe method, pagination will not be used
        We rely on Kafka to get assigned to the partitions and get the offsets
        """
        topics = [topic.name for topic in self.config.topics]
        self.consumer.subscribe(topics)
        t1 = datetime.now()
        encoded_messages = self.consumer.consume(self.config.batch_size, timeout=self.config.consumer_timeout)
        logger.info(f"Kafka consumer read : {len(encoded_messages)} messages in {datetime.now() - t1}")
        records = self.parse_encoded_messages(encoded_messages)
        return SourceIteration(
            next_pagination={},
            records=records,
        )

    def get(self, pagination: dict = None) -> SourceIteration:
        if self.config.sync_mode == SourceSyncModes.STREAM:
            return self.read_topics_with_subscribe(pagination)
        else:
            return self.read_topics_manually(pagination)

    def commit(self):
        self.consumer.commit(asynchronous=False)



import os

from bizon.engine.engine import RunnerFactory

if __name__ == "__main__":
    runner = RunnerFactory.create_from_yaml(filepath=os.path.abspath("test-pipeline-streaming.yml"))
    runner.run()



from datetime import datetime
from enum import Enum
from uuid import uuid4

from pytz import UTC
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import DeclarativeBase, relationship

TABLE_STREAM_INFO = "stream_jobs"
TABLE_SOURCE_CURSOR = "source_cursors"
TABLE_DESTINATION_CURSOR = "destination_cursors"


def generate_uuid():
    return str(uuid4().hex)


class JobStatus(str, Enum):
    STARTED = "started"
    RUNNING = "running"
    FAILED = "failed"
    CANCELED = "cancelled"
    SUCCEEDED = "succeeded"


class CursorStatus(str, Enum):
    STARTED = "started"
    PULLING = "pulling"
    PULLED = "pulled"
    WRITTEN_IN_DESTINATION = "written_in_destination"
    FAILED = "failed"


class Base(DeclarativeBase):
    pass


class StreamJob(Base):
    __tablename__ = TABLE_STREAM_INFO

    id = Column(String(100), primary_key=True, default=generate_uuid, doc="Unique identifier for the job")
    name = Column(String(100), nullable=False, doc="Name of the configuration, must be unique for a given pipeline")
    source_name = Column(String(100), nullable=False, doc="Name of the source")
    stream_name = Column(String(100), nullable=False, doc="Name of the stream")
    sync_mode = Column(String(100), nullable=False, doc="Mode of the sync")
    attempt = Column(Integer, default=0, doc="Number of attempts to run the job")
    total_records_to_fetch = Column(
        Integer, nullable=True, default=None, doc="Total number of records present in the source"
    )
    created_at = Column(
        DateTime, nullable=False, default=datetime.now(tz=UTC), doc="Timestamp when the job was created"
    )
    updated_at = Column(DateTime, nullable=True, default=None, doc="Timestamp when the job was last updated")
    incremental_state = Column(
        String, nullable=True, default=None, doc="Incremental state information from latest sync"
    )
    status = Column(String(100), default=JobStatus.STARTED, doc="Status of the job")
    error_message = Column(String(255), nullable=True, doc="Error message if the job failed", default=None)

    source_cursor = relationship("SourceCursor", cascade="all, delete")
    destination_cursor = relationship("DestinationCursor", cascade="all, delete")

    def __repr__(self):
        return f"<Job {self.id} {self.source_name} {self.stream_name} {self.status}>"


class SourceCursor(Base):
    __tablename__ = TABLE_SOURCE_CURSOR

    id = Column(String(100), primary_key=True, default=generate_uuid, doc="Unique identifier for the cursor")
    job_id = Column(ForeignKey(f"{TABLE_STREAM_INFO}.id"))
    name = Column(String(100), nullable=False, doc="Name of the configuration, must be unique for a given pipeline")
    source_name = Column(String(100), nullable=False, doc="Name of the source")
    stream_name = Column(String(100), nullable=False, doc="Name of the stream")
    iteration = Column(Integer, default=0)
    rows_fetched = Column(Integer, default=0)
    next_pagination = Column(String, nullable=True)
    attempt = Column(Integer, default=0, doc="Number of attempts to pull the data for this cursor")
    status = Column(String(100), default=CursorStatus.STARTED, doc="Status of the cursor")
    error_message = Column(
        String(500), nullable=True, doc="Error message if pulling failed for this cursor", default=None
    )
    created_at = Column(DateTime, default=datetime.now(tz=UTC))
    updated_at = Column(DateTime, nullable=True, default=None, doc="Timestamp when the job was last updated")

    def __repr__(self):
        return f"<SourceCursor {self.stream_name} Iteration: {self.iteration}) - fetched: {self.rows_fetched} at {self.created_at}>"


class DestinationCursor(Base):
    __tablename__ = TABLE_DESTINATION_CURSOR

    id = Column(String(100), primary_key=True, default=generate_uuid, doc="Unique identifier for the cursor")
    job_id = Column(ForeignKey(f"{TABLE_STREAM_INFO}.id"))
    name = Column(String(100), nullable=False, doc="Name of the configuration, must be unique for a given pipeline")
    source_name = Column(String(100), nullable=False, doc="Name of the source")
    stream_name = Column(String(100), nullable=False, doc="Name of the stream")
    destination_name = Column(String(100), nullable=False, doc="Name of the destination")
    from_source_iteration = Column(Integer, default=0)
    to_source_iteration = Column(Integer, default=0)
    rows_written = Column(Integer, default=0)
    success = Column(Boolean, nullable=False, doc="Whether the write operation was successful")
    attempt = Column(Integer, default=0, doc="Number of attempts to pull the data for this cursor")
    error_message = Column(
        String(500), nullable=True, doc="Error message if pulling failed for this cursor", default=None
    )
    created_at = Column(DateTime, default=datetime.now(tz=UTC))
    updated_at = Column(DateTime, nullable=True, default=None, doc="Timestamp when the job was last updated")
    pagination = Column(
        String, nullable=True, default=None, doc="Pagination source information from latest written buffer"
    )



from pydantic import BaseModel, Field

from .abstract_token import AbstractHeaderAuthenticator


class TokenAuthParams(BaseModel):
    token: str = Field(..., description="Token to be attached to the request")
    auth_method: str = Field("Bearer", description="Auth method for token auth")
    auth_header: str = Field("Authorization", description="Auth header for token auth")


class TokenAuthenticator(AbstractHeaderAuthenticator):
    """
    Builds auth header, based on the token provided.
    The token is attached to each request via the `auth_header` header.
    """

    @property
    def auth_header(self) -> str:
        return self._auth_header

    @property
    def token(self) -> str:
        return f"{self._auth_method} {self._token}"

    def __init__(
        self,
        params: TokenAuthParams,
    ):
        self._auth_header = params.auth_header
        self._auth_method = params.auth_method
        self._token = params.token



import multiprocessing
import multiprocessing.synchronize
import threading
import traceback
from abc import ABC, abstractmethod
from typing import Union

from loguru import logger

from bizon.destination.destination import AbstractDestination
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.engine.queue.config import (
    QUEUE_TERMINATION,
    AbstractQueueConfig,
    QueueMessage,
)
from bizon.monitoring.monitor import AbstractMonitor
from bizon.transform.transform import Transform


class AbstractQueueConsumer(ABC):
    def __init__(
        self,
        config: AbstractQueueConfig,
        destination: AbstractDestination,
        transform: Transform,
        monitor: AbstractMonitor,
    ):
        self.config = config
        self.destination = destination
        self.transform = transform
        self.monitor = monitor

    @abstractmethod
    def run(self, stop_event: Union[multiprocessing.synchronize.Event, threading.Event]) -> PipelineReturnStatus:
        pass

    def process_queue_message(self, queue_message: QueueMessage) -> PipelineReturnStatus:

        # Apply the transformation
        try:
            df_source_records = self.transform.apply_transforms(df_source_records=queue_message.df_source_records)
        except Exception as e:
            logger.error(f"Error applying transformation: {e}")
            logger.error(traceback.format_exc())
            self.monitor.track_pipeline_status(PipelineReturnStatus.TRANSFORM_ERROR)
            return PipelineReturnStatus.TRANSFORM_ERROR

        # Handle last iteration
        try:
            if queue_message.signal == QUEUE_TERMINATION:
                logger.info("Received termination signal, waiting for destination to close gracefully ...")
                self.destination.write_records_and_update_cursor(
                    df_source_records=df_source_records,
                    iteration=queue_message.iteration,
                    extracted_at=queue_message.extracted_at,
                    pagination=queue_message.pagination,
                    last_iteration=True,
                )
                self.monitor.track_pipeline_status(PipelineReturnStatus.SUCCESS)
                return PipelineReturnStatus.SUCCESS

        except Exception as e:
            logger.error(f"Error writing records to destination: {e}")
            self.monitor.track_pipeline_status(PipelineReturnStatus.DESTINATION_ERROR)
            return PipelineReturnStatus.DESTINATION_ERROR

        # Write the records to the destination
        try:
            self.destination.write_records_and_update_cursor(
                df_source_records=df_source_records,
                iteration=queue_message.iteration,
                extracted_at=queue_message.extracted_at,
                pagination=queue_message.pagination,
            )
            return PipelineReturnStatus.RUNNING

        except Exception as e:
            logger.error(f"Error writing records to destination: {e}")
            self.monitor.track_pipeline_status(PipelineReturnStatus.DESTINATION_ERROR)
            return PipelineReturnStatus.DESTINATION_ERROR

        raise RuntimeError("Should not reach this point")



from abc import ABC, abstractmethod
from typing import Dict, List

from loguru import logger

from bizon.alerting.models import AlertingConfig, AlertMethod, LogLevel


class AbstractAlert(ABC):

    def __init__(self, type: AlertMethod, config: AlertingConfig, log_levels: List[LogLevel] = [LogLevel.ERROR]):
        self.type = type
        self.config = config
        self.log_levels = log_levels

    @abstractmethod
    def handler(self, message: Dict) -> None:
        pass

    def add_handlers(self) -> None:
        levels = [level.value for level in self.log_levels]
        for level in levels:
            logger.add(self.handler, level=level, format="{message}")



from datetime import datetime
from typing import Optional

from loguru import logger
from pytz import UTC

from bizon.engine.backend.models import JobStatus


class Cursor:
    def __init__(self, source_name: str, stream_name: str, job_id: str, total_records: int = None):

        self.source_name: str = source_name
        self.stream_name: str = stream_name
        self._job_id: str = job_id
        self.iteration: int = 0
        self.rows_fetched: int = 0
        self.total_records = total_records
        self.job_status: JobStatus = JobStatus.STARTED
        self.created_at: datetime = datetime.now(tz=UTC)

        # Instantiate empty pagination
        self._pagination: dict = {}

    @classmethod
    def from_db(
        cls,
        source_name: str,
        stream_name: str,
        job_id: str,
        total_records: int,
        iteration: int,
        rows_fetched: int,
        pagination: dict,
    ):
        cursor = cls(source_name=source_name, stream_name=stream_name, job_id=job_id, total_records=total_records)
        cursor.iteration = iteration
        cursor.rows_fetched = rows_fetched
        cursor._pagination = pagination
        return cursor

    @property
    def job_id(self) -> str:
        return self._job_id

    @property
    def pagination(self) -> dict:
        if self._pagination:
            return self._pagination

        # If it's the first iteration, we return an empty dict
        if self.iteration == 0:
            return dict()

        raise ValueError("Pagination is empty")

    @property
    def is_finished(self) -> bool:
        if self.job_status == JobStatus.SUCCEEDED:
            return True
        return False

    @property
    def percentage_fetched(self) -> Optional[float]:
        """Return the percentage of records fetched from the total records available in the source"""
        if self.total_records is None:
            return None
        return self.rows_fetched / self.total_records

    @property
    def avg_records_per_iteration(self) -> Optional[float]:
        """Return the average number of records fetched per iteration"""
        if self.iteration == 0:
            return None
        return int(self.rows_fetched / self.iteration)

    @property
    def source_full_name(self) -> str:
        return f"{self.source_name}.{self.stream_name}"

    def update_state(self, pagination_dict: dict, nb_records_fetched: int):

        # - 1. update the pagination
        self._pagination = pagination_dict

        # - 2. update the iteration number
        self.iteration += 1

        # - 3. update the number of rows fetched
        self.rows_fetched += nb_records_fetched

        # - 4 Log the progress with humanized percentage
        percentage_str = f"({self.percentage_fetched:.3%})" if self.percentage_fetched else ""

        logger.info(
            (
                f"Source: {self.source_full_name} - Iteration {self.iteration} - "
                f"Fetched: {self.rows_fetched} {percentage_str} successfully."
            )
        )

        # - 5 Handle next status depending on the pagination

        # *** In case pagination is empty, we consider we finished syncing the source ****
        if pagination_dict is None or len(pagination_dict) == 0:
            self.job_status = JobStatus.SUCCEEDED

            if self.total_records is not None and self.rows_fetched != self.total_records:
                logger.info(
                    (
                        f"Source: {self.source_full_name} - Iteration {self.iteration} - "
                        f"Total records fetched: {self.rows_fetched} "
                        f"does not match the total records available: {self.total_records}"
                    )
                )
            else:
                logger.info(
                    (
                        f"Source: {self.source_full_name} - Iteration {self.iteration} - "
                        f"Total records fetched: {self.rows_fetched} "
                        f"matches the total records available: {self.total_records}"
                    )
                )
            return

        # *** Otherwise, we consider the job is still running ***
        self.job_status = JobStatus.RUNNING



import requests


class BaseBackoffException(requests.exceptions.HTTPError):
    def __init__(self, request: requests.PreparedRequest, response: requests.Response, error_message: str = ""):
        error_message = (
            error_message
            or f"Request URL: {request.url}, Response Code: {response.status_code}, Response Text: {response.text}"
        )
        super().__init__(error_message, request=request, response=response)


class DefaultBackoffException(BaseBackoffException):
    pass



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from bizon.alerting.slack.config import SlackConfig


class LogLevel(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class AlertMethod(str, Enum):
    """Alerting methods"""

    SLACK = "slack"


class AlertingConfig(BaseModel):
    """Alerting configuration model"""

    type: AlertMethod
    log_levels: Optional[List[LogLevel]] = [LogLevel.ERROR]
    config: Union[SlackConfig]



from abc import ABC
from enum import Enum

from pydantic import BaseModel, Field


class BackendTypes(str, Enum):
    POSTGRES = "postgres"
    BIGQUERY = "bigquery"

    # For testing purposes or local if no database available
    SQLITE = "sqlite"  # Store in a file bizon.db
    SQLITE_IN_MEMORY = "sqlite_in_memory"  # Only for testing purposes, no persistence


class AbstractBackendConfigDetails(BaseModel, ABC):
    database: str = Field(..., description="Database name")
    schema_name: str = Field(..., description="Schema name", alias="schema")
    syncCursorInDBEvery: int = Field(10, description="Number of iterations before syncing the cursor in the database")


class AbstractBackendConfig(BaseModel, ABC):
    type: BackendTypes = Field(..., description="Type of the backend")
    config: AbstractBackendConfigDetails = Field(..., description="Configuration details for the backend")



import yaml


def parse_from_yaml(path_to_yaml) -> dict:
    with open(path_to_yaml) as f:
        config = yaml.safe_load(f)
    return config


# TODO: Refacto
def set_log_level(config: dict, level: str):
    # Set Log Level to DEBUG
    if level:
        if "engine" not in config:
            config["engine"] = {}

        if "runner" not in config["engine"]:
            config["engine"]["runner"] = {"log_level": level}

        config["engine"]["runner"]["log_level"] = level


def set_custom_source_path_in_config(config: dict, custom_source: str):
    if custom_source:
        config["source"]["source_file_path"] = custom_source


# TODO: Refacto
def set_runner_in_config(config: dict, runner: str):
    if runner:
        if "engine" not in config:
            config["engine"] = {}

        if "runner" not in config["engine"]:
            config["engine"]["runner"] = {"type": runner}

        config["engine"]["runner"]["type"] = runner



import requests
from loguru import logger
from requests.adapters import HTTPAdapter, Retry
from requests.exceptions import HTTPError
from requests.models import Response


# Define a named function instead of using a lambda to be able to pickle it
def raise_for_status_hook(response: Response, *args, **kwargs):
    response.raise_for_status()


class Session(requests.Session):
    def __init__(self):
        super().__init__()

        # call response.raise_for_status() after every response
        self.hooks["response"] = [raise_for_status_hook]

        # Implement basic rety policy if rate-limited
        retries = Retry(
            total=20,
            backoff_factor=1,
            raise_on_status=True,
            status=30,
            allowed_methods=["GET", "POST"],
        )

        self.mount(
            "https://",
            HTTPAdapter(
                max_retries=retries,
                pool_maxsize=64,
            ),
        )
        self._method_mapping = {
            "POST": self.post,
            "GET": self.get,
        }

    def call(
        self,
        method: str,
        url: str,
        content_type: str = "application/json",
        *args,
        **kwargs,
    ) -> requests.Response:

        self.headers.update({"content-type": content_type})

        try:
            response = self._method_mapping[method](url=url, *args, **kwargs)
        except HTTPError as e:
            logger.error(f"Error {e}")
            logger.error(f"detailed error response: {e.response.json()}")
            logger.error(f"for request body: {e.request.body}")
            raise e
        except Exception as e:
            logger.error(f"Error {e}")
            logger.error(
                f"""for request '{method}' on url='{url}'
                              with args={args}
                              and kwargs={kwargs}
                          """
            )
            raise e
        else:
            return response



import os
import time
from datetime import datetime
from typing import List

import polars as pl
import simplejson as json
from loguru import logger
from pytz import UTC

from bizon.common.models import BizonConfig
from bizon.destination.models import transform_to_df_destination_records
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.engine.runner.config import RunnerStatus
from bizon.engine.runner.runner import AbstractRunner
from bizon.source.models import SourceRecord, source_record_schema


class StreamingRunner(AbstractRunner):
    def __init__(self, config: BizonConfig):
        super().__init__(config)

    @staticmethod
    def convert_source_records(records: List[SourceRecord]) -> pl.DataFrame:
        return pl.DataFrame(
            {
                "id": [record.id for record in records],
                "data": [json.dumps(record.data, ensure_ascii=False) for record in records],
                "timestamp": [record.timestamp for record in records],
                "destination_id": [record.destination_id for record in records],
            },
            schema=source_record_schema,
        )

    @staticmethod
    def convert_to_destination_records(df_source_records: pl.DataFrame, extracted_at: datetime) -> pl.DataFrame:
        return transform_to_df_destination_records(df_source_records=df_source_records, extracted_at=extracted_at)

    def run(self) -> RunnerStatus:
        job = self.init_job(bizon_config=self.bizon_config, config=self.config)
        backend = self.get_backend(bizon_config=self.bizon_config)
        source = self.get_source(bizon_config=self.bizon_config, config=self.config)
        destination = self.get_destination(
            bizon_config=self.bizon_config,
            backend=backend,
            job_id=job.id,
            source_callback=None,
        )
        transform = self.get_transform(bizon_config=self.bizon_config)
        monitor = self.get_monitoring_client(bizon_config=self.bizon_config)
        destination.buffer.buffer_size = 0  # force buffer to be flushed immediately
        iteration = 0

        while True:

            if source.config.max_iterations and iteration > source.config.max_iterations:
                logger.info(f"Max iterations {source.config.max_iterations} reached, terminating stream ...")
                break

            source_iteration = source.get()

            destination_id_indexed_records = {}

            if len(source_iteration.records) == 0:
                logger.info("No new records found, stopping iteration")
                time.sleep(2)
                monitor.track_pipeline_status(PipelineReturnStatus.SUCCESS)
                iteration += 1
                continue

            for record in source_iteration.records:
                if destination_id_indexed_records.get(record.destination_id):
                    destination_id_indexed_records[record.destination_id].append(record)
                else:
                    destination_id_indexed_records[record.destination_id] = [record]

            for destination_id, records in destination_id_indexed_records.items():
                df_source_records = StreamingRunner.convert_source_records(records)

                # Apply transformation
                df_source_records = transform.apply_transforms(df_source_records=df_source_records)

                df_destination_records = StreamingRunner.convert_to_destination_records(
                    df_source_records, datetime.now(tz=UTC)
                )
                # Override destination_id
                destination.destination_id = destination_id
                destination.write_or_buffer_records(
                    df_destination_records=df_destination_records,
                    iteration=iteration,
                    pagination=None,
                )
                monitor.track_records_synced(
                    num_records=len(df_destination_records),
                    extra_tags={"destination_id": destination_id},
                )
            if os.getenv("ENVIRONMENT") == "production":
                source.commit()

            iteration += 1

            monitor.track_pipeline_status(PipelineReturnStatus.SUCCESS)
        return RunnerStatus(stream=PipelineReturnStatus.SUCCESS)  # return when max iterations is reached



import os
from typing import Dict

from datadog import initialize, statsd
from loguru import logger

from bizon.common.models import BizonConfig
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.monitoring.monitor import AbstractMonitor


class DatadogMonitor(AbstractMonitor):
    def __init__(self, pipeline_config: BizonConfig):
        super().__init__(pipeline_config)

        # In Kubernetes, set the host dynamically
        try:
            datadog_host_from_env_var = os.getenv(pipeline_config.monitoring.config.datadog_host_env_var)
            if datadog_host_from_env_var:
                initialize(
                    statsd_host=datadog_host_from_env_var,
                    statsd_port=pipeline_config.monitoring.config.datadog_agent_port,
                )
            else:
                initialize(
                    statsd_host=pipeline_config.monitoring.config.datadog_agent_host,
                    statsd_port=pipeline_config.monitoring.config.datadog_agent_port,
                )
        except Exception as e:
            logger.info(f"Failed to initialize Datadog agent: {e}")

        self.pipeline_monitor_status = "bizon_pipeline.status"
        self.tags = [
            f"pipeline_name:{self.pipeline_config.name}",
            f"pipeline_stream:{self.pipeline_config.source.stream}",
            f"pipeline_source:{self.pipeline_config.source.name}",
            f"pipeline_destination:{self.pipeline_config.destination.name}",
        ] + [f"{key}:{value}" for key, value in self.pipeline_config.monitoring.config.tags.items()]

        self.pipeline_active_pipelines = "bizon_pipeline.active_pipelines"
        self.pipeline_records_synced = "bizon_pipeline.records_synced"

    def track_pipeline_status(self, pipeline_status: PipelineReturnStatus, extra_tags: Dict[str, str] = {}) -> None:
        """
        Track the status of the pipeline.

        Args:
            status (str): The current status of the pipeline (e.g., 'running', 'failed', 'completed').
        """

        statsd.increment(
            self.pipeline_monitor_status,
            tags=self.tags
            + [f"pipeline_status:{pipeline_status}"]
            + [f"{key}:{value}" for key, value in extra_tags.items()],
        )

    def track_records_synced(self, num_records: int, extra_tags: Dict[str, str] = {}) -> None:
        """
        Track the number of records synced in the pipeline.

        Args:
            num_records (int): Number of records synced in this batch
        """
        statsd.increment(
            self.pipeline_records_synced,
            value=num_records,
            tags=self.tags + [f"{key}:{value}" for key, value in extra_tags.items()],
        )



from abc import ABC, abstractmethod
from typing import Dict

from bizon.common.models import BizonConfig
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.monitoring.config import MonitorType


class AbstractMonitor(ABC):
    def __init__(self, pipeline_config: BizonConfig):
        self.pipeline_config = pipeline_config
        # Initialize the monitor

    @abstractmethod
    def track_pipeline_status(self, pipeline_status: PipelineReturnStatus, extra_tags: Dict[str, str] = {}) -> None:
        """
        Track the status of the pipeline.

        Args:
            status (str): The current status of the pipeline (e.g., 'running', 'failed', 'completed').
        """
        pass

    def track_records_synced(self, num_records: int, extra_tags: Dict[str, str] = {}) -> None:
        """
        Track the number of records synced in the pipeline.
        """
        pass


class MonitorFactory:
    @staticmethod
    def get_monitor(pipeline_config: BizonConfig) -> AbstractMonitor:
        if pipeline_config.monitoring is None:
            from bizon.monitoring.noop.monitor import NoOpMonitor

            return NoOpMonitor(pipeline_config)

        if pipeline_config.monitoring.type == MonitorType.DATADOG:
            from bizon.monitoring.datadog.monitor import DatadogMonitor

            return DatadogMonitor(pipeline_config)



from typing import Any, List, Mapping, Union

import pendulum
from pydantic import BaseModel, Field
from pydantic_extra_types.pendulum_dt import DateTime

from .abstract_oauth import AbstractOauth2Authenticator


class Oauth2AuthParams(BaseModel):
    token_refresh_endpoint: str = Field(..., description="URL to refresh the token")
    client_id: str = Field(..., description="Client ID")
    client_secret: str = Field(..., description="Client Secret")
    refresh_token: str = Field(..., description="Refresh Token")
    scopes: List[str] = Field(None, description="Scopes")
    token_expiry_date: DateTime = Field(None, description="Token expiry date")
    token_expiry_date_format: str = Field(None, description="Token expiry date format")
    access_token_name: str = Field("access_token", description="Name of the access token")
    expires_in_name: str = Field("expires_in", description="Name of the expires in")
    refresh_request_body: Mapping[str, Any] = Field(None, description="Request body to refresh the token")
    grant_type: str = Field("refresh_token", description="Grant type")


class Oauth2Authenticator(AbstractOauth2Authenticator):
    """
    Generates OAuth2.0 access tokens from an OAuth2.0 refresh token and client credentials.
    The generated access token is attached to each request via the Authorization header.
    """

    def __init__(self, params: Oauth2AuthParams):
        self._token_refresh_endpoint = params.token_refresh_endpoint
        self._client_secret = params.client_secret
        self._client_id = params.client_id
        self._refresh_token = params.refresh_token
        self._scopes = params.scopes
        self._access_token_name = params.access_token_name
        self._expires_in_name = params.expires_in_name
        self._refresh_request_body = params.refresh_request_body
        self._grant_type = params.grant_type

        self._token_expiry_date = params.token_expiry_date or pendulum.now().subtract(days=1)
        self._token_expiry_date_format = params.token_expiry_date_format
        self._access_token = None

    def get_token_refresh_endpoint(self) -> str:
        return self._token_refresh_endpoint

    def get_client_id(self) -> str:
        return self._client_id

    def get_client_secret(self) -> str:
        return self._client_secret

    def get_refresh_token(self) -> str:
        return self._refresh_token

    def get_access_token_name(self) -> str:
        return self._access_token_name

    def get_scopes(self) -> List[str]:
        return self._scopes

    def get_expires_in_name(self) -> str:
        return self._expires_in_name

    def get_refresh_request_body(self) -> Mapping[str, Any]:
        return self._refresh_request_body

    def get_grant_type(self) -> str:
        return self._grant_type

    def get_token_expiry_date(self) -> DateTime:
        return self._token_expiry_date

    def set_token_expiry_date(self, value: Union[str, int]):
        if self._token_expiry_date_format:
            self._token_expiry_date = pendulum.from_format(value, self._token_expiry_date_format)
        else:
            self._token_expiry_date = pendulum.now().add(seconds=value)

    @property
    def access_token(self) -> str:
        return self._access_token

    @access_token.setter
    def access_token(self, value: str):
        self._access_token = value



from typing import Union

import pika
from loguru import logger

from bizon.destination.destination import AbstractDestination
from bizon.engine.pipeline.consumer import AbstractQueueConsumer
from bizon.engine.queue.config import QUEUE_TERMINATION, QueueMessage
from bizon.engine.queue.queue import AbstractQueue

from .config import RabbitMQConfigDetails
from .consumer import RabbitMQConsumer


class RabbitMQ(AbstractQueue):

    def __init__(self, config: RabbitMQConfigDetails) -> None:
        super().__init__(config)
        self.config: RabbitMQConfigDetails = config

    def connect(self):
        self.connection = pika.BlockingConnection(pika.ConnectionParameters(host=self.config.queue.host))
        channel = self.connection.channel()
        channel.queue_declare(queue=self.config.queue.queue_name)
        self.channel = channel

    def get_consumer(self, destination: AbstractDestination) -> AbstractQueueConsumer:
        return RabbitMQConsumer(config=self.config, destination=destination)

    def put_queue_message(self, queue_message: QueueMessage):
        self.channel.basic_publish(
            exchange=self.config.queue.exchange,
            routing_key=self.config.queue.queue_name,
            body=queue_message.model_dump_json(),
        )

    def get_size(self) -> Union[int, None]:
        return None

    def get(self) -> QueueMessage:
        raise NotImplementedError(
            "RabbitMQ does not support getting messages from the queue, directly use callback in consumer."
        )

    def terminate(self, iteration: int) -> bool:
        self.put(source_records=[], iteration=iteration, signal=QUEUE_TERMINATION)
        logger.info("Sent termination signal to destination.")
        return True



import os
from typing import Dict, List

import requests
from loguru import logger

from bizon.alerting.alerts import AbstractAlert, AlertMethod
from bizon.alerting.models import LogLevel
from bizon.alerting.slack.config import SlackConfig


class SlackHandler(AbstractAlert):
    def __init__(self, config: SlackConfig, log_levels: List[LogLevel] = [LogLevel.ERROR]):
        super().__init__(type=AlertMethod.SLACK, config=config, log_levels=log_levels)
        self.webhook_url = config.webhook_url

    def handler(self, message: Dict) -> None:
        """
        Custom handler to send error logs to Slack, with additional context.
        """
        log_entry = message.record
        error_message = (
            f"*Sync*: `{os.environ.get('BIZON_SYNC_NAME', 'N/A')}`\n"
            f"*Source*: `{os.environ.get('BIZON_SOURCE_NAME', 'N/A')}` - `{os.environ.get('BIZON_SOURCE_STREAM', 'N/A')}`\n"  # noqa
            f"*Destination*: `{os.environ.get('BIZON_DESTINATION_NAME', 'N/A')}`\n\n"
            f"*Message:*\n```{log_entry['message']}```\n"
            f"*File:* `{log_entry['file'].path}:{log_entry['line']}`\n"
            f"*Function:* `{log_entry['function']}`\n"
            f"*Level:* `{log_entry['level'].name}`\n"
        )

        payload = {"text": f":rotating_light: *Bizon Pipeline Alert* :rotating_light:\n\n{error_message}"}

        try:
            response = requests.post(self.webhook_url, json=payload)
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed to send log to Slack: {e}")
        return None



import pika
import pika.connection
from loguru import logger

from bizon.destination.destination import AbstractDestination
from bizon.engine.queue.config import QUEUE_TERMINATION
from bizon.engine.queue.queue import AbstractQueueConsumer, QueueMessage

from .config import RabbitMQConfigDetails


class RabbitMQConsumer(AbstractQueueConsumer):
    def __init__(self, config: RabbitMQConfigDetails, destination: AbstractDestination):
        super().__init__(config, destination=destination)
        self.config: RabbitMQConfigDetails = config

    def run(self) -> None:
        connection = pika.BlockingConnection(
            pika.ConnectionParameters(host=self.config.queue.host, port=self.config.queue.port)
        )

        channel = connection.channel()

        channel.queue_declare(queue=self.config.queue.queue_name)

        for method_frame, properties, body in channel.consume(self.config.queue.queue_name):

            queue_message = QueueMessage.model_validate_json(body)
            if queue_message.signal == QUEUE_TERMINATION:
                logger.info("Received termination signal, waiting for destination to close gracefully ...")
                self.destination.write_records_and_update_cursor(
                    source_records=queue_message.source_records,
                    iteration=queue_message.iteration,
                    extracted_at=queue_message.extracted_at,
                    pagination=queue_message.pagination,
                    last_iteration=True,
                )
                channel.queue_delete(queue=self.config.queue.queue_name)
                channel.close()
                break

            self.destination.write_records_and_update_cursor(
                source_records=queue_message.source_records,
                iteration=queue_message.iteration,
                extracted_at=queue_message.extracted_at,
                pagination=queue_message.pagination,
            )
            logger.info(f"Consumed data from queue: {len(queue_message.source_records)}")



from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class HubSpotProperty(BaseModel):
    name: str
    label: str
    field_type: str
    type: str
    description: Optional[str]


class HubSpotObjectProperty(BaseModel):
    value: Any


class AllObjectProperties(BaseModel):
    properties: List[HubSpotProperty]

    @property
    def names(self) -> List[str]:
        return [hs_property.name for hs_property in self.properties]

    def property_names(
        self,
    ) -> List[str]:
        return [hs_property.name for hs_property in self.properties]


class HubSpotObject(BaseModel):
    id: int
    properties: Dict[str, HubSpotObjectProperty]
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_raw_obj_response(
        cls,
        raw_obj: dict,
    ):

        properties = {}

        for property_name, property_value in raw_obj.get("properties", {}).items():
            properties[property_name] = HubSpotObjectProperty(value=property_value)

        return cls(
            id=raw_obj["id"],
            properties=properties,
            created_at=raw_obj["createdAt"],
            updated_at=raw_obj["updatedAt"],
        )



import random
from typing import List, Literal, Tuple, Union

from pydantic import Field
from requests.auth import AuthBase

from bizon.source.auth.authenticators.oauth import Oauth2AuthParams
from bizon.source.auth.authenticators.token import TokenAuthParams
from bizon.source.auth.builder import AuthBuilder
from bizon.source.auth.config import AuthConfig, AuthType
from bizon.source.config import SourceConfig, SourceSyncModes
from bizon.source.models import SourceIteration, SourceRecord
from bizon.source.source import AbstractSource

from .fake_api import fake_api_call

URL_TOKEN_REFRESH = "https://api.mydummysaas.com/oauth/v1/token"


class DummyAuthConfig(AuthConfig):
    type: Literal[AuthType.API_KEY, AuthType.OAUTH]
    params: Union[TokenAuthParams, Oauth2AuthParams] = Field(None, description="OAuth or API configuration")


class DummySourceConfig(SourceConfig):
    authentication: DummyAuthConfig
    sleep: int = Field(0, description="Sleep time in seconds between API calls")


class DummySource(AbstractSource):

    def __init__(self, config: DummySourceConfig):
        super().__init__(config)
        self.config = config

    @staticmethod
    def streams() -> List[str]:
        return ["creatures", "plants"]

    @staticmethod
    def get_config_class() -> SourceConfig:
        return DummySourceConfig

    @property
    def url_entity(self) -> str:
        return f"https://api.dummy.com/v1/{self.config.stream}"

    def get_authenticator(self) -> AuthBase:

        if self.config.authentication.type == AuthType.OAUTH:
            return AuthBuilder.oauth2(
                params=Oauth2AuthParams(
                    token_refresh_endpoint=URL_TOKEN_REFRESH,
                    client_id=self.config.authentication.params.client_id,
                )
            )

        if self.config.authentication.type == AuthType.API_KEY:
            return AuthBuilder.token(params=TokenAuthParams(token=self.config.authentication.params.token))

    def check_connection(self) -> Tuple[bool, str | None]:
        # Here we could check if the connection is established
        return True, None

    def get_total_records_count(self) -> int | None:
        # If available, return total number of records in the source for this stream
        # Otherwise, return None
        # In our case we have 5 records
        return 5

    def get(self, pagination: dict = None) -> SourceIteration:

        response: dict = None

        # If no pagination data is passed, we want to reach first page
        if not pagination:
            response = fake_api_call(url=self.url_entity, sleep=self.config.sleep)

        # If we have pagination data we pass it to the API
        else:
            response = fake_api_call(url=self.url_entity, cursor=pagination.get("cursor"), sleep=self.config.sleep)

        # Now we process the response to:
        # - allow bizon to process the records and write them to destination
        # - iterate on next page if needed

        # We parse records and next cursor from the response
        records = response.get("results")
        next_cursor = response.get("next", {}).get("cursor")

        next_pagination = {"cursor": next_cursor} if next_cursor else {}

        destination_id = None

        # If we are in streaming mode, we need to get the destination id from the stream name
        if self.config.sync_mode == SourceSyncModes.STREAM:
            if next_pagination.get("cursor") == "final-cursor":
                destination_id = "routed"
            else:
                destination_id = self.config.stream

        if records:
            return SourceIteration(
                next_pagination=next_pagination,
                records=[
                    SourceRecord(
                        id=record["id"],
                        data=record,
                        destination_id=destination_id,
                    )
                    for record in records
                ],
            )

        return SourceIteration(
            next_pagination=next_pagination,
            records=[],
        )



from enum import Enum
from typing import Any, List, Literal, Mapping

from pydantic import BaseModel, Field

from bizon.source.auth.config import AuthConfig, AuthType
from bizon.source.config import SourceConfig


class SchemaRegistryType(str, Enum):
    APICURIO = "apicurio"


class MessageEncoding(str, Enum):
    UTF_8 = "utf-8"
    AVRO = "avro"


class KafkaAuthConfig(AuthConfig):

    type: Literal[AuthType.BASIC] = AuthType.BASIC  # username and password authentication

    # Schema registry authentication
    schema_registry_type: SchemaRegistryType = Field(
        default=SchemaRegistryType.APICURIO, description="Schema registry type"
    )

    schema_registry_url: str = Field(default="", description="Schema registry URL with the format ")
    schema_registry_username: str = Field(default="", description="Schema registry username")
    schema_registry_password: str = Field(default="", description="Schema registry password")


def default_kafka_consumer_config():
    return {
        "auto.offset.reset": "earliest",
        "enable.auto.commit": False,  # Turn off auto-commit for manual offset handling
        "session.timeout.ms": 45000,
        "security.protocol": "SASL_SSL",
    }


class TopicConfig(BaseModel):
    name: str = Field(..., description="Kafka topic name")
    destination_id: str = Field(..., description="Destination id")


class KafkaSourceConfig(SourceConfig):

    # Mandatory Kafka configuration
    topics: List[TopicConfig] = Field(..., description="Kafka topic, comma separated")
    bootstrap_servers: str = Field(..., description="Kafka bootstrap servers")
    group_id: str = Field(default="bizon", description="Kafka group id")

    skip_message_empty_value: bool = Field(
        default=True, description="Skip messages with empty value (tombstone messages)"
    )

    # Kafka consumer configuration
    batch_size: int = Field(100, description="Kafka batch size, number of messages to fetch at once.")
    consumer_timeout: int = Field(10, description="Kafka consumer timeout in seconds, before returning batch.")

    consumer_config: Mapping[str, Any] = Field(
        default_factory=default_kafka_consumer_config,
        description="Kafka consumer configuration, as described in the confluent-kafka-python documentation",
    )

    message_encoding: str = Field(default=MessageEncoding.AVRO, description="Encoding to use to decode the message")

    # Schema ID header configuration
    nb_bytes_schema_id: Literal[4, 8] = Field(
        description="Number of bytes encode SchemaID in Kafka message. Standard is 4.",
        default=4,
    )

    authentication: KafkaAuthConfig = Field(..., description="Authentication configuration")



from typing import Literal

from pydantic import BaseModel, Field

from bizon.engine.queue.config import (
    AbastractQueueConfigDetails,
    AbstractQueueConfig,
    QueueTypes,
)


class RabbitMQQueueConfig(BaseModel):
    host: str = Field(..., description="RabbitMQ host")
    port: int = Field(5672, description="RabbitMQ port")
    queue_name: str = Field(..., description="RabbitMQ queue")
    exchange: str = Field("", description="RabbitMQ exchange")


class RabbitMQConsumerConfig(BaseModel):
    poll_interval: int = Field(1, description="Interval in seconds to poll the queue in seconds")


class RabbitMQConfigDetails(AbastractQueueConfigDetails):
    queue: RabbitMQQueueConfig = Field(..., description="RabbitMQ queue configuration")
    consumer: RabbitMQConsumerConfig = Field(RabbitMQConsumerConfig(), description="Rabbitmq consumer configuration")


class RabbitMQConfig(AbstractQueueConfig):
    type: Literal[QueueTypes.RABBITMQ]
    config: RabbitMQConfigDetails



from .authenticators.basic import BasicHttpAuthenticator, BasicHttpAuthParams
from .authenticators.cookies import CookiesAuthenticator, CookiesAuthParams
from .authenticators.oauth import Oauth2Authenticator, Oauth2AuthParams
from .authenticators.token import TokenAuthenticator, TokenAuthParams


class AuthBuilder:
    @staticmethod
    def oauth2(params: Oauth2AuthParams) -> Oauth2Authenticator:
        return Oauth2Authenticator(params=params)

    @staticmethod
    def basic(params: BasicHttpAuthParams) -> BasicHttpAuthenticator:
        return BasicHttpAuthenticator(params=params)

    @staticmethod
    def token(params: TokenAuthParams) -> TokenAuthenticator:
        return TokenAuthenticator(params=params)

    @staticmethod
    def cookies(params: CookiesAuthParams) -> CookiesAuthenticator:
        return CookiesAuthenticator(params=params)



import json

from kafka import KafkaConsumer
from loguru import logger

from bizon.destination.destination import AbstractDestination
from bizon.engine.pipeline.consumer import AbstractQueueConsumer
from bizon.engine.queue.config import QUEUE_TERMINATION, QueueMessage

from .config import KafkaConfigDetails


class KafkaConsumer_(AbstractQueueConsumer):
    def __init__(self, config: KafkaConfigDetails, destination: AbstractDestination):
        super().__init__(config, destination=destination)
        self.config: KafkaConfigDetails = config
        self.consumer = self.get_consumer()
        self.consumer.subscribe(self.config.queue.topic)

    def get_consumer(self) -> KafkaConsumer:
        return KafkaConsumer(
            bootstrap_servers=self.config.queue.bootstrap_server,
            group_id=self.config.consumer.group_id,
            auto_offset_reset=self.config.consumer.auto_offset_reset,
            enable_auto_commit=self.config.consumer.enable_auto_commit,
            consumer_timeout_ms=self.config.consumer.consumer_timeout_ms,
            value_deserializer=lambda m: json.loads(m.decode("utf-8")),
        )

    def run(self):
        try:
            for message in self.consumer:
                logger.debug(f"Consuming message on topic: {message.partition}|{message.offset} key: {message.key}")
                queue_message = QueueMessage.model_validate(message.value)

                if queue_message.signal == QUEUE_TERMINATION:
                    logger.info("Received termination signal, waiting for destination to close gracefully ...")
                    self.destination.write_records_and_update_cursor(
                        source_records=queue_message.source_records,
                        iteration=queue_message.iteration,
                        extracted_at=queue_message.extracted_at,
                        pagination=queue_message.pagination,
                        last_iteration=True,
                    )
                    break

                self.destination.write_records_and_update_cursor(
                    source_records=queue_message.source_records,
                    extracted_at=queue_message.extracted_at,
                    iteration=queue_message.iteration,
                    pagination=queue_message.pagination,
                )

                logger.info(f"Consumed data from queue: {len(queue_message.source_records)}")

        except Exception as e:
            logger.error(f"Error occurred while consuming messages: {e}")
        finally:
            self.consumer.close()



from yaml import safe_load

from bizon.engine.engine import RunnerFactory

config_yaml = """
source:
  name: dummy
  stream: creatures
  authentication:
    type: api_key
    params:
      token: dummy_key

destination:
  # Authentication: If empty it will be infered.
  # Must have the bigquery.jobUser
  # Must have the bigquery.dataEditor and storage.objectUser on the supplied dataset and bucket
  name: bigquery
  config:
    dataset_id: bizon_test
    dataset_location: US
    project_id: <project_id>
    gcs_buffer_bucket: bizon-buffer
    gcs_buffer_format: parquet
"""
config = safe_load(config_yaml)
runner = RunnerFactory.create_from_config_dict(config=config)
runner.run()



from typing import Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from bizon.alerting.models import AlertingConfig
from bizon.connectors.destinations.bigquery.src.config import BigQueryConfig
from bizon.connectors.destinations.bigquery_streaming.src.config import (
    BigQueryStreamingConfig,
)
from bizon.connectors.destinations.bigquery_streaming_v2.src.config import (
    BigQueryStreamingV2Config,
)
from bizon.connectors.destinations.file.src.config import FileDestinationConfig
from bizon.connectors.destinations.logger.src.config import LoggerConfig
from bizon.engine.config import EngineConfig
from bizon.monitoring.config import MonitoringConfig
from bizon.source.config import SourceConfig, SourceSyncModes
from bizon.transform.config import TransformModel


class BizonConfig(BaseModel):

    # Forbid extra keys in the model
    model_config = ConfigDict(extra="forbid")

    # Unique name to identify the sync configuration
    name: str = Field(..., description="Unique name for this sync configuration")

    source: SourceConfig = Field(
        description="Source configuration",
        default=...,
    )

    transforms: Optional[list[TransformModel]] = Field(
        description="List of transformations to apply to the source data",
        default=[],
    )

    destination: Union[
        BigQueryConfig,
        BigQueryStreamingConfig,
        BigQueryStreamingV2Config,
        LoggerConfig,
        FileDestinationConfig,
    ] = Field(
        description="Destination configuration",
        discriminator="name",
        default=...,
    )

    engine: EngineConfig = Field(
        description="Engine configuration",
        default=EngineConfig(),
    )

    alerting: Optional[AlertingConfig] = Field(
        description="Alerting configuration",
        default=None,
    )

    monitoring: Optional[MonitoringConfig] = Field(
        description="Monitoring configuration",
        default=None,
    )


class SyncMetadata(BaseModel):
    """Model which stores general metadata around a sync.
    Facilitate usage of basic info across entities
    """

    name: str
    job_id: str
    source_name: str
    stream_name: str
    sync_mode: SourceSyncModes
    destination_name: str

    @classmethod
    def from_bizon_config(cls, job_id: str, config: BizonConfig) -> "SyncMetadata":
        return cls(
            name=config.name,
            job_id=job_id,
            source_name=config.source.name,
            stream_name=config.source.stream,
            sync_mode=config.source.sync_mode,
            destination_name=config.destination.name,
        )



from typing import List

from bizon.source.callback import AbstractSourceCallback
from bizon.source.models import SourceIteration

from .config import KafkaSourceConfig


class KafkaSourceCallback(AbstractSourceCallback):
    def __init__(self, config: KafkaSourceConfig):
        super().__init__(config)

    def on_iterations_written(self, iterations: List[SourceIteration]):
        """Commit the offsets of the iterations"""

        # TODO: Implement the callback

        pass



from abc import ABC
from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field

from .auth.config import AuthConfig


class SourceSyncModes(str, Enum):

    # Full refresh
    # - creates a new StreamJob
    # - start syncing data from scratch till pagination is empty
    FULL_REFRESH = "full_refresh"

    # Incremental syncs data
    # - get the last successful sync start_date from previous succeeded StreamJob
    # - create a new StreamJob
    # - sync data till pagination is empty
    INCREMENTAL = "incremental"

    # Stream mode
    # - get single RUNNING streamJob syncs data from the last successfuly committed offset in the source system
    STREAM = "stream"


class APIConfig(BaseModel):
    retry_limit: Optional[int] = Field(100, description="Number of retries before giving up", example=100)


class SourceConfig(BaseModel, ABC):

    # Connector identifier to use, match a unique connector code
    name: str = Field(..., description="Name of the source to sync")
    stream: str = Field(..., description="Name of the stream to sync")

    source_file_path: Optional[str] = Field(
        default=None, description="Path to the source file, if not provided will look into bizon internal sources"
    )

    sync_mode: SourceSyncModes = Field(
        description="Sync mode to use",
        default=SourceSyncModes.FULL_REFRESH,
    )

    force_ignore_checkpoint: bool = Field(
        description="Whether to force recreate the sync from iteration 0. Existing checkpoints will be ignored.",
        default=False,
    )

    authentication: Optional[AuthConfig] = Field(
        description="Configuration for the authentication",
        default=None,
    )

    max_iterations: Optional[int] = Field(
        description="Maximum number of iterations for pulling from source, if None, run till all records are synced from source",  # noqa
        default=None,
    )

    api_config: APIConfig = Field(
        description="Configuration for the API client",
        default=APIConfig(retry_limit=10),
    )

    init_pipeline: bool = Field(
        description="[Used for testing] Whether to initialize the source to run pipeline directly.",
        default=True,
    )



import io
import os
import tempfile
import traceback
from typing import List, Tuple
from uuid import uuid4

import polars as pl
from google.api_core.exceptions import NotFound
from google.cloud import bigquery, storage
from google.cloud.bigquery import DatasetReference, TimePartitioning
from loguru import logger

from bizon.common.models import SyncMetadata
from bizon.destination.destination import AbstractDestination
from bizon.engine.backend.backend import AbstractBackend
from bizon.source.config import SourceSyncModes
from bizon.source.source import AbstractSourceCallback

from .config import BigQueryColumn, BigQueryConfigDetails


class BigQueryDestination(AbstractDestination):

    def __init__(
        self,
        sync_metadata: SyncMetadata,
        config: BigQueryConfigDetails,
        backend: AbstractBackend,
        source_callback: AbstractSourceCallback,
    ):
        super().__init__(sync_metadata, config, backend, source_callback)
        self.config: BigQueryConfigDetails = config

        if config.authentication and config.authentication.service_account_key:
            with tempfile.NamedTemporaryFile(delete=False) as temp:
                temp.write(config.authentication.service_account_key.encode())
                temp_file_path = temp.name
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = temp_file_path

        self.project_id = config.project_id
        self.bq_client = bigquery.Client(project=self.project_id)
        self.gcs_client = storage.Client(project=self.project_id)
        self.buffer_bucket_name = config.gcs_buffer_bucket
        self.buffer_bucket = self.gcs_client.bucket(config.gcs_buffer_bucket)
        self.buffer_format = config.gcs_buffer_format
        self.dataset_id = config.dataset_id
        self.dataset_location = config.dataset_location

    @property
    def table_id(self) -> str:
        tabled_id = self.destination_id or f"{self.sync_metadata.source_name}_{self.sync_metadata.stream_name}"
        return f"{self.project_id}.{self.dataset_id}.{tabled_id}"

    @property
    def temp_table_id(self) -> str:

        if self.sync_metadata.sync_mode == SourceSyncModes.FULL_REFRESH:
            return f"{self.table_id}_temp"

        elif self.sync_metadata.sync_mode == SourceSyncModes.INCREMENTAL:
            return f"{self.table_id}_incremental"

        elif self.sync_metadata.sync_mode == SourceSyncModes.STREAM:
            return f"{self.table_id}"

    def get_bigquery_schema(self, df_destination_records: pl.DataFrame) -> List[bigquery.SchemaField]:

        # Case we unnest the data
        if self.config.unnest:
            return [
                bigquery.SchemaField(
                    col.name,
                    col.type,
                    mode=col.mode,
                    description=col.description,
                )
                for col in self.record_schemas[self.destination_id]
            ]

        # Case we don't unnest the data
        else:
            return [
                bigquery.SchemaField("_source_record_id", "STRING", mode="REQUIRED"),
                bigquery.SchemaField("_source_timestamp", "TIMESTAMP", mode="REQUIRED"),
                bigquery.SchemaField("_source_data", "STRING", mode="NULLABLE"),
                bigquery.SchemaField("_bizon_extracted_at", "TIMESTAMP", mode="REQUIRED"),
                bigquery.SchemaField(
                    "_bizon_loaded_at", "TIMESTAMP", mode="REQUIRED", default_value_expression="CURRENT_TIMESTAMP()"
                ),
                bigquery.SchemaField("_bizon_id", "STRING", mode="REQUIRED"),
            ]

    def check_connection(self) -> bool:
        dataset_ref = DatasetReference(self.project_id, self.dataset_id)

        try:
            self.bq_client.get_dataset(dataset_ref)
        except NotFound:
            dataset = bigquery.Dataset(dataset_ref)
            dataset.location = self.dataset_location
            dataset = self.bq_client.create_dataset(dataset)
        return True

    def cleanup(self, gcs_file: str):
        blob = self.buffer_bucket.blob(gcs_file)
        blob.delete()

    # TO DO: Add backoff to common exceptions => looks like most are hanlded by the client
    # https://cloud.google.com/python/docs/reference/storage/latest/retry_timeout
    # https://cloud.google.com/python/docs/reference/bigquery/latest/google.cloud.bigquery.dbapi.DataError

    def convert_and_upload_to_buffer(self, df_destination_records: pl.DataFrame) -> str:

        if self.buffer_format == "parquet":

            # Upload the Parquet file to GCS
            file_name = f"{self.sync_metadata.source_name}/{self.sync_metadata.stream_name}/{str(uuid4())}.parquet"

            with io.BytesIO() as stream:
                df_destination_records.write_parquet(stream)
                stream.seek(0)

                blob = self.buffer_bucket.blob(file_name)
                blob.upload_from_file(stream, content_type="application/octet-stream")

            return file_name

        raise NotImplementedError(f"Buffer format {self.buffer_format} is not supported")

    @staticmethod
    def unnest_data(df_destination_records: pl.DataFrame, record_schema: list[BigQueryColumn]) -> pl.DataFrame:
        """Unnest the source_data field into separate columns"""

        # Check if the schema matches the expected schema
        source_data_fields = (
            pl.DataFrame(df_destination_records["source_data"].str.json_decode(infer_schema_length=None))
            .schema["source_data"]
            .fields
        )

        record_schema_fields = [col.name for col in record_schema]

        for field in source_data_fields:
            assert field.name in record_schema_fields, f"Column {field.name} not found in BigQuery schema"

        # Parse the JSON and unnest the fields to polar type
        return df_destination_records.select(
            pl.col("source_data").str.json_path_match(f"$.{col.name}").cast(col.polars_type).alias(col.name)
            for col in record_schema
        )

    def load_to_bigquery(self, gcs_file: str, df_destination_records: pl.DataFrame):

        # We always partition by the loaded_at field
        time_partitioning = TimePartitioning(field="_bizon_loaded_at", type_=self.config.time_partitioning)

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.PARQUET,
            write_disposition=bigquery.WriteDisposition.WRITE_APPEND,
            schema=self.get_bigquery_schema(df_destination_records=df_destination_records),
            time_partitioning=time_partitioning,
        )

        load_job = self.bq_client.load_table_from_uri(
            f"gs://{self.buffer_bucket_name}/{gcs_file}", self.temp_table_id, job_config=job_config
        )
        result = load_job.result()  # Waits for the job to complete
        assert result.state == "DONE", f"Job failed with state {result.state} with error {result.error_result}"

    def write_records(self, df_destination_records: pl.DataFrame) -> Tuple[bool, str]:

        # Rename fields to match BigQuery schema
        df_destination_records = df_destination_records.rename(
            {
                # Bizon fields
                "bizon_extracted_at": "_bizon_extracted_at",
                "bizon_id": "_bizon_id",
                "bizon_loaded_at": "_bizon_loaded_at",
                # Source fields
                "source_record_id": "_source_record_id",
                "source_timestamp": "_source_timestamp",
                "source_data": "_source_data",
            },
        )

        gs_file_name = self.convert_and_upload_to_buffer(df_destination_records=df_destination_records)

        try:
            self.load_to_bigquery(gcs_file=gs_file_name, df_destination_records=df_destination_records)
            self.cleanup(gs_file_name)
        except Exception as e:
            self.cleanup(gs_file_name)
            logger.error(f"Error loading data to BigQuery: {e}")
            logger.error(traceback.format_exc())
            return False, str(e)
        return True, ""

    def finalize(self):
        if self.sync_metadata.sync_mode == SourceSyncModes.FULL_REFRESH:
            logger.info(f"Loading temp table {self.temp_table_id} data into {self.table_id} ...")
            self.bq_client.query(f"CREATE OR REPLACE TABLE {self.table_id} AS SELECT * FROM {self.temp_table_id}")
            logger.info(f"Deleting temp table {self.temp_table_id} ...")
            self.bq_client.delete_table(self.temp_table_id, not_found_ok=True)
            return True

        elif self.sync_metadata.sync_mode == SourceSyncModes.INCREMENTAL:
            # TO DO: Implement incremental sync
            return True

        elif self.sync_metadata.sync_mode == SourceSyncModes.STREAM:
            # Nothing to do as we write directly to the final table
            return True



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple, Type, Union

from requests.auth import AuthBase

from .callback import AbstractSourceCallback, NoOpSourceCallback
from .config import SourceConfig
from .models import SourceIncrementalState, SourceIteration, SourceRecord
from .session import Session


class AbstractSource(ABC):

    def __init__(self, config: SourceConfig):
        self.config = config
        self.session = self.get_session()

        # Set authentication in the session
        auth = self.get_authenticator()

        if auth:
            self.session.auth = auth

    def __del__(self):
        self.session.close()

    @staticmethod
    @abstractmethod
    def streams() -> List[str]:
        """Return all the streams that the source supports"""
        pass

    @staticmethod
    @abstractmethod
    def get_config_class() -> Type[SourceConfig]:
        """Return the config class for the source"""
        pass

    def get_source_callback_instance(self) -> AbstractSourceCallback:
        """Return an instance of the source callback"""
        return NoOpSourceCallback(config=self.config)

    @abstractmethod
    def get_authenticator(self) -> Union[AuthBase, None]:
        """Return an authenticator for the source, it will be set in the session
        If no authenticator is needed, return None
        """
        pass

    @abstractmethod
    def check_connection(self) -> Tuple[bool, Optional[Any]]:
        pass

    @abstractmethod
    def get(self, pagination: dict = None) -> SourceIteration:
        """Perform next API call to retrieve data and return next pagination.

        If pagination is None, it should return the first page of data.
        Otherwise, it should return the next page of data.

        If no pagination is returned in SourceIteration, it means that there is no more data to fetch.

        - pagination dict
        - records List[dict]
        """
        pass

    @abstractmethod
    def get_total_records_count(self) -> Optional[int]:
        """Return the total count of records available in the source"""
        pass

    def get_records_after(self, source_state: SourceIncrementalState, pagination: dict = None) -> SourceIteration:
        """Perform next API call to retrieve data incrementally and return next pagination.

        If pagination is None, it should return the first page of the incremental data.

        Otherwise, it should return the next page of data.

        If no pagination is returned in SourceIteration, it means that:
          - there is no more data to fetch
          -  you need to provide the next state that will be stored.
        """

        # In case we need a special formatting for the date or a non-date state format
        # We can use the `state` dict in SourceIncrementalState

        # In case of HubSpot we only use last_run timestamp from state
        # -> Iterate on all records updated between source_state.last_run and now
        # -> Return the pagination of the search API
        # -> perform new search if needed (more than 10K results from CRM Search API)

        # In case of Kafka
        # -> Iterate on all records after last offset stored in state
        # -> Return the next offset to fetch in pagination to continue syncing data
        # -> When no records are in the topic for a given time (timeout), we consider there is no data to fetch
        # -> return empty pagination and return last offset pulled to put in the state
        pass

    def get_session(self) -> Session:
        """Return a new session"""
        return Session()

    def commit(self, records: Optional[List[SourceRecord]] = None):
        """Commit the records to the source"""
        pass



import random
import time
from multiprocessing import Queue
from typing import Union

from loguru import logger

from bizon.destination.destination import AbstractDestination
from bizon.engine.queue.config import QUEUE_TERMINATION, QueueMessage
from bizon.engine.queue.queue import AbstractQueue, AbstractQueueConsumer
from bizon.monitoring.monitor import AbstractMonitor
from bizon.source.callback import AbstractSourceCallback
from bizon.source.models import SourceIteration
from bizon.transform.transform import Transform

from .config import PythonQueueConfigDetails
from .consumer import PythonQueueConsumer


class PythonQueue(AbstractQueue):

    def __init__(self, config: PythonQueueConfigDetails, **kwargs) -> None:
        super().__init__(config)
        self.config: PythonQueueConfigDetails = config

        assert "queue" in kwargs, "queue param passed in kwargs is required for PythonQueue"
        self.queue: Queue = kwargs["queue"]

    def connect(self):
        # No connection to establish for PythonQueue
        pass

    def get_consumer(
        self,
        destination: AbstractDestination,
        transform: Transform,
        monitor: AbstractMonitor,
    ) -> AbstractQueueConsumer:
        return PythonQueueConsumer(
            config=self.config,
            queue=self.queue,
            destination=destination,
            transform=transform,
            monitor=monitor,
        )

    def put_queue_message(self, queue_message: QueueMessage):
        if not self.queue.full():
            self.queue.put(queue_message)
            logger.debug(f"Putting data from iteration {queue_message.iteration} items in queue)")
        else:
            logger.warning("Queue is full, waiting for consumer to consume data")
            time.sleep(random.random())
            self.put_queue_message(queue_message)

    def get(self) -> QueueMessage:
        if not self.queue.empty():
            queue_message: QueueMessage = self.queue.get()
            logger.debug(f"Got {len(queue_message.df_source_records.height)} records from queue")
            return queue_message
        else:
            logger.debug("Queue is empty, waiting for producer to produce data")
            time.sleep(random.random())
            return self.get()

    def get_size(self) -> Union[int, None]:
        if hasattr(self.queue, "qsize"):
            return self.queue.qsize()
        return None

    def terminate(self, iteration: int) -> bool:
        self.put(
            source_iteration=SourceIteration(next_pagination={}, records=[]),
            iteration=iteration,
            signal=QUEUE_TERMINATION,
        )
        logger.info("Sent termination signal to destination.")
        return True



from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Optional

import polars as pl
from pydantic import BaseModel, ConfigDict, Field
from pytz import UTC

QUEUE_TERMINATION = "TERMINATION"


@dataclass
class QueueMessage:
    iteration: int
    df_source_records: pl.DataFrame
    extracted_at: datetime = datetime.now(tz=UTC)
    pagination: Optional[dict] = None
    signal: Optional[str] = None


class QueueTypes(str, Enum):
    KAFKA = "kafka"
    RABBITMQ = "rabbitmq"
    PYTHON_QUEUE = "python_queue"


class AbastractQueueConfigDetails(BaseModel, ABC):

    # Forbid extra keys in the model
    model_config = ConfigDict(extra="forbid")

    max_nb_messages: int = Field(1_000_000, description="Maximum number of messages in the queue")

    queue: BaseModel = Field(..., description="Configuration of the queue")
    consumer: BaseModel = Field(..., description="Configuration of the consumer")


class AbstractQueueConfig(BaseModel, ABC):

    # Forbid extra keys in the model
    model_config = ConfigDict(extra="forbid")

    type: QueueTypes = Field(..., description="Type of the queue")
    config: AbastractQueueConfigDetails = Field(..., description="Configuration of the queue")



from abc import ABC
from http import HTTPStatus
from typing import Any, List, Optional, Tuple

from requests import HTTPError
from requests.adapters import HTTPAdapter, Retry
from requests.auth import AuthBase

from bizon.source.auth.authenticators.oauth import Oauth2AuthParams
from bizon.source.auth.authenticators.token import TokenAuthParams
from bizon.source.auth.builder import AuthBuilder
from bizon.source.auth.config import AuthType
from bizon.source.session import Session
from bizon.source.source import AbstractSource

URL_BASE = "https://api.hubapi.com"
URL_GRANTED_SCOPES = f"{URL_BASE}/oauth/v1/access-tokens"
URL_TOKEN_REFRESH = f"{URL_BASE}/oauth/v1/token"


class HubSpotBaseSource(AbstractSource, ABC):

    def get_session(self) -> Session:
        """Apply custom strategy for HubSpot"""
        session = Session()

        # Retry policy if rate-limited by HubSpot
        retries = Retry(
            total=50,
            backoff_factor=1,
            raise_on_status=True,
            status_forcelist=[429, 500, 502, 503, 504],
            status=30,
            allowed_methods=["GET", "POST"],
        )
        session.mount("https://", HTTPAdapter(max_retries=retries, pool_maxsize=64))
        return session

    def get_authenticator(self) -> AuthBase:
        if self.config.authentication.type.value == AuthType.OAUTH.value:
            return AuthBuilder.oauth2(
                params=Oauth2AuthParams(
                    token_refresh_endpoint=URL_TOKEN_REFRESH,
                    client_id=self.config.authentication.params.client_id,
                    client_secret=self.config.authentication.params.client_secret,
                    refresh_token=self.config.authentication.params.refresh_token,
                )
            )

        elif self.config.authentication.type.value == AuthType.API_KEY.value:
            return AuthBuilder.token(
                params=TokenAuthParams(
                    token=self.config.authentication.params.token,
                )
            )

        raise NotImplementedError(f"Auth type {self.config.authentication.type} not implemented for HubSpot")

    def check_connection(self) -> Tuple[bool, Optional[Any]]:
        """Check connection"""
        alive = True
        error_msg = None
        try:
            objects, state = self.get()
        except HTTPError as error:
            alive = False
            error_msg = repr(error)
            if error.response.status_code == HTTPStatus.BAD_REQUEST:
                response_json = error.response.json()
                error_msg = (
                    f"400 Bad Request: {response_json['message']}, please check if provided credentials are valid."
                )
        except Exception as e:
            alive = False
            error_msg = repr(e)
        return alive, error_msg

    def get_granted_scopes(self) -> List[str]:
        try:
            if self.config.authentication.type == AuthType.OAUTH.value:
                response = self.session.get(url=f"{URL_GRANTED_SCOPES}/{self.session.auth.access_token}")
            else:
                raise NotImplementedError("Scope endpoint for API Key are not supported.")
            response.raise_for_status()
            response_json = response.json()
            granted_scopes = response_json["scopes"]
            return granted_scopes
        except Exception as e:
            return False, repr(e)



import ast
import multiprocessing
import multiprocessing.synchronize
import threading
import traceback
from datetime import datetime
from time import sleep
from typing import Tuple, Union

from loguru import logger
from pytz import UTC

from bizon.common.models import BizonConfig
from bizon.engine.backend.backend import AbstractBackend
from bizon.engine.backend.models import CursorStatus
from bizon.engine.queue.queue import AbstractQueue
from bizon.source.cursor import Cursor
from bizon.source.source import AbstractSource

from .models import PipelineReturnStatus


class Producer:
    def __init__(
        self, bizon_config: BizonConfig, queue: AbstractQueue, source: AbstractSource, backend: AbstractBackend
    ):
        self.bizon_config = bizon_config
        self.queue = queue
        self.source = source
        self.backend = backend

    @property
    def name(self) -> str:
        return f"producer-{self.source.config.name}-{self.source.config.stream}"

    def get_or_create_cursor(self, job_id: str, session=None) -> Cursor:
        """Get or create a cursor for the current stream, return the cursor"""
        # Try to get the cursor from the DB
        cursor_from_db = self.backend.get_last_cursor_by_job_id(job_id=job_id)

        cursor = None

        if cursor_from_db:
            # Retrieve the job
            job = self.backend.get_stream_job_by_id(job_id=job_id)

            logger.info(
                f"Recovered cursor from DB for iteration {cursor_from_db.from_source_iteration} -> {cursor_from_db.to_source_iteration}"
            )

            # Initialize the recovery from the DestinationCursor
            cursor = Cursor.from_db(
                source_name=self.source.config.name,
                stream_name=self.source.config.stream,
                job_id=job_id,
                total_records=job.total_records_to_fetch,
                iteration=cursor_from_db.to_source_iteration + 1,
                rows_fetched=self.backend.get_number_of_written_rows_for_job(job_id=job_id),
                pagination=ast.literal_eval(cursor_from_db.pagination),
            )
        else:
            # Get the total number of records
            total_records = self.source.get_total_records_count()
            # Initialize the cursor
            cursor = Cursor(
                source_name=self.source.config.name,
                stream_name=self.source.config.stream,
                job_id=job_id,
                total_records=total_records,
            )
        return cursor

    def handle_max_iterations(self, cursor: Cursor) -> bool:
        """Handle the case where the max_iterations is reached for the current cursor
        If max_iterations is reached return True
        Else return False
        """
        if self.source.config.max_iterations and cursor.iteration > self.source.config.max_iterations:
            logger.warning(
                f"Max iteration of {self.source.config.max_iterations} reached for this cursor, terminating ..."
            )
            return True
        return False

    def is_queue_full(self, cursor: Cursor) -> Tuple[bool, int, int]:
        """Handle the case where the queue is full
        If the queue is full return True, 0, 0
        Else return False, int[queue_size], int[approximate_nb_records_in_queue]
        """
        queue_size = self.queue.get_size()

        if not isinstance(queue_size, int):
            return False, 0, 0

        if not isinstance(cursor.avg_records_per_iteration, int):
            return False, 0, 0

        approximate_nb_records_in_queue = queue_size * cursor.avg_records_per_iteration

        if approximate_nb_records_in_queue >= self.queue.config.max_nb_messages:
            return True, queue_size, approximate_nb_records_in_queue

        return False, queue_size, approximate_nb_records_in_queue

    def run(
        self, job_id: int, stop_event: Union[multiprocessing.synchronize.Event, threading.Event]
    ) -> PipelineReturnStatus:

        return_value: PipelineReturnStatus = PipelineReturnStatus.SUCCESS

        # Init queue
        try:
            self.queue.connect()
        except Exception as e:
            logger.error(
                f"Error while connecting to the queue: {e} for job_id {job_id}"
                f"Check the queue error logs for more details."
            )
            return PipelineReturnStatus.QUEUE_ERROR

        # Get or create the cursor
        try:
            cursor = self.get_or_create_cursor(job_id=job_id, session=self.backend.session)
        except Exception as e:
            logger.error(traceback.format_exc())
            logger.error(
                f"Error while getting or creating cursor: {e} for job_id {job_id}"
                f"Check the backend error logs for more details."
            )
            logger.info("Terminating destination ...")
            self.queue.terminate(iteration=0)
            return PipelineReturnStatus.BACKEND_ERROR

        while not cursor.is_finished:

            if stop_event.is_set():
                logger.info("Stop event is set, terminating producer ...")
                return PipelineReturnStatus.KILLED_BY_RUNNER

            timestamp_start_iteration = datetime.now(tz=UTC)

            # Handle the case where last cursor already reach max_iterations
            terminate = self.handle_max_iterations(cursor)
            if terminate:
                break

            # Check if we need to create a new cursor in the DB
            if cursor.iteration % self.backend.config.syncCursorInDBEvery == 0:
                # Create a new cursor in the DB
                try:
                    self.backend.create_source_cursor(
                        job_id=job_id,
                        name=self.bizon_config.name,
                        source_name=self.source.config.name,
                        stream_name=self.source.config.stream,
                        iteration=cursor.iteration,
                        rows_fetched=cursor.rows_fetched,
                        next_pagination=cursor.pagination,
                        cursor_status=CursorStatus.PULLING,
                    )
                except Exception as e:
                    logger.error(traceback.format_exc())
                    logger.error(
                        f"Error while creating cursor in the DB: {e} for iteration {cursor.iteration}"
                        f"Check the backend error logs for more details."
                    )
                    return_value = PipelineReturnStatus.BACKEND_ERROR
                    break

            # Get current time, we consider this as the time the data was extracted
            extracted_at = datetime.now(tz=UTC)

            # Check if queue is full, we wait a random time and retry
            is_queue_full, queue_size, approximate_nb_records_in_queue = self.is_queue_full(cursor)
            if is_queue_full:
                QUEUE_FULL_WAITING_TIME = 2
                logger.warning(
                    f"Queue is full contains {queue_size} iterations with {approximate_nb_records_in_queue} records, waiting {QUEUE_FULL_WAITING_TIME} seconds before retrying..."
                )
                sleep(QUEUE_FULL_WAITING_TIME)
                continue

            # Get the next data
            try:
                source_iteration = self.source.get(pagination=cursor.pagination)
            except Exception as e:
                logger.error(traceback.format_exc())
                logger.error(
                    f"Error while fetching data from source: {e} for iteration first iteration {cursor.iteration}"
                    f"Check the backend error logs for more details."
                )
                return_value = PipelineReturnStatus.SOURCE_ERROR
                break

            # Put the data in the queue
            try:
                self.queue.put(
                    source_iteration=source_iteration,
                    iteration=cursor.iteration,
                    extracted_at=extracted_at,
                )
            except Exception as e:
                logger.error(traceback.format_exc())
                logger.error(
                    f"Error while putting data in the queue: {e} for iteration {cursor.iteration}"
                    f"Check the queue error logs for more details."
                )
                return_value = PipelineReturnStatus.QUEUE_ERROR
                break

            # Update the cursor state
            try:
                cursor.update_state(
                    pagination_dict=source_iteration.next_pagination, nb_records_fetched=len(source_iteration.records)
                )
            except Exception as e:
                logger.error(traceback.format_exc())
                logger.error(
                    f"Error while updating cursor state: {e} for iteration {cursor.iteration}"
                    f"Check the source error logs for more details."
                )
                return_value = PipelineReturnStatus.SOURCE_ERROR
                break

            # Items in queue
            items_in_queue = f"{self.queue.get_size()} items in queue." if self.queue.get_size() else ""

            logger.info(
                (
                    f"Iteration {cursor.iteration} finished in {datetime.now(tz=UTC) - timestamp_start_iteration}. {items_in_queue}"
                )
            )

        logger.info("Terminating destination ...")

        try:
            self.queue.terminate(iteration=cursor.iteration)
        except Exception as e:
            logger.error(traceback.format_exc())
            logger.error(
                f"Error while terminating the queue: {e} for iteration {cursor.iteration}"
                f"Check the queue error logs for more details."
            )
            # Prevent masking any previous error
            return_value = (
                PipelineReturnStatus.QUEUE_ERROR if return_value == PipelineReturnStatus.SUCCESS else return_value
            )

        return return_value



import json
import textwrap

import polars as pl
from loguru import logger

from .config import TransformModel


class Transform:

    def __init__(self, transforms: list[TransformModel]):
        self.transforms = transforms

    def apply_transforms(self, df_source_records: pl.DataFrame) -> pl.DataFrame:
        """Apply transformation on df_source_records"""

        # Process the transformations
        for transform in self.transforms:

            logger.debug(f"Applying transform {transform.label}")

            # Create a function to be executed in the desired context
            def my_transform(data: str) -> str:

                data = json.loads(data)

                # Start writing here
                local_vars = {"data": data}

                # Normalize the indentation of the Python code
                normalized_python = textwrap.dedent(transform.python)

                exec(normalized_python, {}, local_vars)

                # Stop writing here
                return json.dumps(local_vars["data"])

            transformed_source_records = []

            for row in df_source_records["data"].to_list():
                transformed_source_records.append(my_transform(row))

            df_source_records = df_source_records.with_columns(
                pl.Series("data", transformed_source_records, dtype=pl.String).alias("data")
            )

        return df_source_records



from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class FailureType(str, Enum):
    SYSTEM_ERROR = "SYSTEM_ERROR"
    CONFIG_ERROR = "CONFIG_ERROR"


class ErrorTraceMessage(BaseModel):
    model_config = ConfigDict(extra="forbid")

    message: str = Field(..., description="A user-friendly message that indicates the cause of the error")
    internal_message: Optional[str] = Field(None, description="The internal error that caused the failure")
    stack_trace: Optional[str] = Field(None, description="The full stack trace of the error")
    failure_type: Optional[FailureType] = Field(None, description="The type of error")



import io
import struct
from functools import lru_cache

import fastavro
import orjson
from avro.schema import Schema


class Hashabledict(dict):
    def __hash__(self):
        return hash(frozenset(self))


def get_header_bytes(nb_bytes_schema_id: int, message: bytes) -> bytes:
    """Get the header bytes from the message"""

    if nb_bytes_schema_id == 8:
        return message[:9]

    elif nb_bytes_schema_id == 4:
        return message[1:5]

    else:
        raise ValueError(f"Number of bytes for schema id {nb_bytes_schema_id} not supported")


@lru_cache(maxsize=None)
def parse_global_id_from_serialized_message(nb_bytes_schema_id: int, header_message_bytes: bytes) -> int:
    """Parse the global id from the serialized message"""

    if nb_bytes_schema_id == 8:
        return struct.unpack(">bq", header_message_bytes)[1]

    elif nb_bytes_schema_id == 4:
        return struct.unpack(">I", header_message_bytes)[0]

    raise ValueError(f"Number of bytes for schema id {nb_bytes_schema_id} not supported")


@lru_cache(maxsize=None)
def find_debezium_json_fields(hashable_schema: Hashabledict) -> list[str]:
    """Find the JSON fields in the Debezium schema"""

    json_fields = []

    for field in hashable_schema["fields"]:
        if field["name"] == "before":
            for before_types in field["type"]:
                if isinstance(before_types, dict) and before_types["type"] == "record":
                    debezium_columns = before_types["fields"]
                    for column in debezium_columns:
                        if (
                            isinstance(column.get("type"), dict)
                            and column.get("type").get("connect.name") == "io.debezium.data.Json"
                        ):
                            json_fields.append(column["name"])
    return json_fields


def parse_debezium_json_fields(data: dict, hashable_schema: Hashabledict) -> None:
    """Parse the JSON fields from the Debezium payload data in-place"""

    json_fields = find_debezium_json_fields(hashable_schema)

    for field in json_fields:
        for root_column in ["before", "after"]:
            if data.get(root_column) and data.get(root_column).get(field) is not None:
                # We JSON loads the field only if it's not empty, otherwise we return None
                data[root_column][field] = (
                    orjson.loads(data[root_column][field]) if len(data[root_column][field]) > 0 else None
                )


def decode_avro_message(
    message: bytes, nb_bytes_schema_id: int, hashable_dict_schema: Hashabledict, avro_schema: Schema
) -> dict:
    """Decode an Avro message"""

    # Decode the message
    message_bytes = io.BytesIO(message.value())
    message_bytes.seek(nb_bytes_schema_id + 1)
    data = fastavro.schemaless_reader(message_bytes, avro_schema.to_json())

    # Parse the JSON fields in-place
    parse_debezium_json_fields(data=data, hashable_schema=hashable_dict_schema)

    return data
